/* buildmap_main.c - The main function of the map builder tool.
 *
 * LICENSE:
 *
 *   Copyright 2002 Pascal F. Martin
 *
 *   This file is part of RoadMap.
 *
 *   RoadMap is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   RoadMap is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with RoadMap; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>

#include "roadmap_types.h"
#include "roadmap_hash.h"

#include "buildmap.h"
#include "buildmap_tiger.h"

#include "buildmap_square.h"
#include "buildmap_street.h"
#include "buildmap_range.h"
#include "buildmap_line.h"
#include "buildmap_point.h"
#include "buildmap_shape.h"
#include "buildmap_city.h"
#include "buildmap_zip.h"
#include "buildmap_area.h"
#include "buildmap_polygon.h"


static void buildmap_county_initialize (void) {

   buildmap_zip_initialize();
   buildmap_city_initialize();
   buildmap_point_initialize();
   buildmap_range_initialize();
   buildmap_line_initialize();
   buildmap_polygon_initialize();
   buildmap_shape_initialize();
   buildmap_street_initialize();
   buildmap_area_initialize();
}


static void buildmap_county_save (char *name) {

   char *cursor;
   char db_name[128];

   snprintf (db_name, 127, "usc%s", name);

   cursor = strrchr (db_name, '.');
   if (cursor != NULL) {
      *cursor = 0;
   }

   buildmap_db_open (db_name);

   buildmap_square_save ();
   buildmap_line_save ();
   buildmap_point_save ();
   buildmap_shape_save ();
   buildmap_dictionary_save ();
   buildmap_city_save ();
   buildmap_street_save ();
   buildmap_range_save ();
   buildmap_polygon_save ();
   buildmap_zip_save ();

   buildmap_db_close ();
}

static void buildmap_county_reset (void) {

   buildmap_square_reset ();
   buildmap_line_reset ();
   buildmap_point_reset ();
   buildmap_shape_reset ();
   buildmap_dictionary_reset ();
   buildmap_city_reset ();
   buildmap_street_reset ();
   buildmap_range_reset ();
   buildmap_polygon_reset ();
   buildmap_zip_reset ();
   roadmap_hash_reset ();
}

static void buildmap_county_process
               (char *path, char *name, int verbose, int canals, int rivers) {

   buildmap_county_initialize ();

   buildmap_tiger_initialize (verbose, canals, rivers);

   buildmap_tiger_read_rtc (path, name, verbose);

   buildmap_tiger_read_rt7 (path, name, verbose);

   buildmap_tiger_read_rt8 (path, name, verbose);

   buildmap_tiger_read_rti (path, name, verbose);

   buildmap_tiger_read_rt1 (path, name, verbose);

   buildmap_tiger_read_rt2 (path, name, verbose);

   buildmap_tiger_sort();

   if (verbose) {
      roadmap_hash_summary ();
      buildmap_dictionary_summary ();
      buildmap_tiger_summary ();
   }

   buildmap_county_save (name);
   buildmap_county_reset ();
}


int main (int argc, char **argv) {

   int i;
   int verbose = 0;
   int canals = 0;
   int rivers = 0;
   char *argument = NULL;
   char *extension;
   char *path = NULL;
   DIR  *directory;
   struct dirent *entry;


   for (i = 1; i < argc; i++) {

      argument = argv[i];

      if (argument[0] != '-') break;

      if (strcmp (argument, "--help") == 0) {

         fprintf (stderr, "Usage: %s [-v] [path]\n", argv[0]);
         exit(0);

      } else if (strcmp (argument, "--canal") == 0) {

         canals = 1;

      } else if (strcmp (argument, "--river") == 0) {

         rivers = 1;

      } else if (argument[1] == 'v' && argument[2] == 0) {

         verbose = 1;

      } else {
         fprintf (stderr, "%s: bad argument %s\n", argv[0], argument);
         return 1;
      }
   }

   if (i < argc) {
      path = argument;
      i++;
   } else {
      path = ".";
   }

   directory = opendir (path);

   if (directory == NULL) {
      fprintf (stderr, "cannot open directory %s\n", path);
   }

   if (i < argc) {

      for (; i < argc; i++) {
         buildmap_county_process (path, argv[i], verbose, canals, rivers);
         if (verbose) {
            printf ("Done with county %s\n", argv[i]);
         }
      }

   } else {

      for (entry = readdir (directory);
           entry != NULL;
           entry = readdir (directory)) {

         extension = entry->d_name + strlen(entry->d_name) - 4;

         if (extension == NULL) continue;

         if (strncmp (entry->d_name, "TGR", 3) == 0 &&
             strcmp (extension, ".RT1") == 0) {

            char *from;
            char *to;
            char  county[8];

            for (from = entry->d_name + 3, to = county;
                 from < extension && to < county + 7;
                 from++, to++) {
               *to = *from;
            }
            *to = 0;

            buildmap_county_process (path, county, verbose, canals, rivers);
            if (verbose) {
               printf ("Done with county %s\n", county);
            }
         }
      }
   }

   closedir (directory);

   return 0;
}

