/* buildus_fips.c - Read the list of FIPS codes from the Census Bureau files.
 *
 * LICENSE:
 *
 *   Copyright 2002 Pascal F. Martin
 *
 *   This file is part of RoadMap.
 *
 *   RoadMap is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   RoadMap is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with RoadMap; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * SYNOPSYS:
 *
 *   void buildus_fips_read (char *path, int verbose);
 *   void buildus_fips_summary (void);
 */

#define _ISOC9X_SOURCE

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <ctype.h>

#include "roadmap_types.h"
#include "roadmap_hash.h"

#include "buildmap.h"
#include "buildus_fips.h"
#include "buildus_county.h"


static FILE *buildus_fips_open
                (char *path, char *name, char *extension, int verbose) {

   FILE *file;

   char full_name[1025];

   snprintf (full_name, 1024, "%s/%s%s", path, name, extension);

   buildmap_set_source (full_name);

   if (verbose) {

      struct stat state_result;

      if (stat (full_name, &state_result) != 0) {
         buildmap_fatal (0, "cannot stat file");
      }

      buildmap_info ("size = %d Kbytes", state_result.st_size / 1024);
   }

   file = fopen (full_name, "r");
   if (file == NULL) {
      buildmap_fatal (0, "cannot open file %s", full_name);
   }

   return file;
}


void buildus_fips_read (char *path, int verbose) {

   FILE *input;

   static BuildMapDictionary state_dictionary;
   static BuildMapDictionary county_dictionary;

   char *p;
   char *q;
   char  line[1024];
   int   line_count;

   int fips;
   int level;
   RoadMapString state_name;
   RoadMapString state_symbol;
   RoadMapString county_name;


   input = buildus_fips_open (path, "AllSt", ".txt", verbose);

   state_dictionary = buildmap_dictionary_open ("state");
   county_dictionary = buildmap_dictionary_open ("county");

   line_count = 0;

   while (! feof(input)) {

      if (fgets (line, sizeof(line), input) == NULL) {
         break;
      }

      line_count += 1;
      buildmap_set_line (line_count);

      fips = 0;
      level = 0;

      /* Retrieve and decode the FIPS code. */

      for (p = line; isblank(*p); p++) ;

      if (isalpha(*p)) {
         goto next_line; /* This is a comment line. */
      }

      if (*p < ' ') {
         goto next_line; /* This is an empty line. */
      }

      for (p = line; *p > 0; p++) {

         if (isdigit(*p)) {

            for (q = p; isdigit(*q); q++) ;

            if (q - p == 4) {

               /* This is of the county/area code we skip. */
               p = q;
               continue;
            }

            if (q - p == 5) {

               /* This is the fips code we was looking for. */
               fips = atoi(p);

               p = q;
               break;
            }

            printf ("** invalid line: %s", line);
            goto next_line; /* Not a line we are interested on. */
         }

         if (! isblank(*p)) {
            printf ("** invalid line: %s", line);
            goto next_line; /* Not a line we are interested on. */
         }
      }

      /* retrieve and decode the level code. */

      for (; *p > 0; p++) {

         if (isdigit(*p)) {

            if (! isblank(p[1])) {
               printf ("** invalid line: %s", line);
               goto next_line; /* Not a line we are interested on. */
            }

            level = (*p - '0');

            if (level == 1) {
               if (fips %1000 != 0) {
                  buildmap_error (p - line, "invalid state FIPS: %s", line);
                  goto next_line; /* Not a line we are interested on. */
               }
            } else if (level <= 0 || level > 5) {
               buildmap_error (p - line, "invalid level: %s", line);
               goto next_line; /* Not a line we are interested on. */
            }

            p += 1; /* Skip the level. */
            break;
         }

         if (! isblank(*p)) {

            printf ("** invalid character '%c' after FIPS: %s", *p, line);
            goto next_line; /* Not a line we are interested on. */
         }
      }

      if ((fips == 0) || (level == 0)) {
         buildmap_fatal (p - line, "invalid decoding");
      }

      /* retrieve and decode the name. */

      while (isblank(*p) && *p > 0) p++;

      if (! isalpha(*p)) {
         printf ("** invalid character '%c' in name: %s", *p, line);
         goto next_line; /* Not a line we are interested on. */
      }

      q = strchr(p, ',');

      if (level == 1) {

         if (q != NULL) {

            printf ("** invalid state name: %s", line);
            goto next_line; /* Not a line we are interested on. */
         }

         for (q = p; *q >= ' '; q++) ;

         *q = 0;
         state_name = buildmap_dictionary_add (state_dictionary, p, q - p);

         buildus_county_add_state (state_name, fips / 1000);

      } else {

         if (q == NULL) {

            if (strncasecmp (p, "district of columbia", 20) != 0) {
               printf ("** invalid county name: %s", line);
               goto next_line; /* Not a line we are interested on. */
            }

            for (q = p; *q >= ' '; q++) ;

            *q = 0;
            county_name = buildmap_dictionary_add (county_dictionary, p, q - p);
            state_symbol = buildmap_dictionary_add (state_dictionary, "DC", 2);

         } else {

            *q = 0;
            county_name = buildmap_dictionary_add (county_dictionary, p, q - p);

            do {
               q++;
            } while (isblank(*q));

            if (! isalpha(*q)) {
               printf ("** invalid state acronym: %s", line);
               goto next_line; /* Not a line we are interested on. */
            }

            p = q;
            for (q = p; *q >= ' '; q++) ;

            *q = 0;
            state_symbol = buildmap_dictionary_add (state_dictionary, p, q - p);
         }

         buildus_county_add (fips, county_name, state_symbol);
      }

      next_line:
   }

   fclose(input);
}

void buildus_fips_summary (void) {
}

