/* roadmap_canvas.c - manage the canvas that is used to draw the map.
 *
 * LICENSE:
 *
 *   Copyright 2002 Pascal F. Martin
 *
 *   This file is part of RoadMap.
 *
 *   RoadMap is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   RoadMap is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with RoadMap; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * SYNOPSYS:
 *
 *   See roadmap_canvas.h.
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "roadmap.h"
#include "roadmap_types.h"
#include "roadmap_gui.h"
#include "roadmap_math.h"
#include "roadmap_square.h"
#include "roadmap_street.h"
#include "roadmap_line.h"
#include "roadmap_shape.h"
#include "roadmap_point.h"
#include "roadmap_polygon.h"
#include "roadmap_dictionary.h"
#include "roadmap_county.h"
#include "roadmap_locator.h"


#define ROADMAP_CANVAS_POINT_BLOCK  1024

#define ROADMAP_DECLUTTER_STREET     160
#define ROADMAP_DECLUTTER_MAIN      1300
#define ROADMAP_DECLUTTER_FREEWAY 0x7fffffff

#define ROADMAP_DECLUTTER_SEA     0x1000
#define ROADMAP_DECLUTTER_LAKE      1300
#define ROADMAP_DECLUTTER_RIVER      300

#define ROADMAP_THICKNESS_STREET    1
#define ROADMAP_THICKNESS_MAIN      2
#define ROADMAP_THICKNESS_FREEWAY   3

#define ROADMAP_THICKNESS_SHORE     2
#define ROADMAP_THICKNESS_RIVER     3

#define ROADMAP_DECLUTTER_POLYGONS 0x1000
#define ROADMAP_DECLUTTER_SHAPE     1300

#define ROADMAP_SEGMENT_BLOCK 256

#define ROADMAP_CURSOR_SIZE   10


typedef struct roadmap_canvas_sprite {

   RoadMapPosition position;

   GdkGC *pen;

} RoadMapSprite;

static GdkGC        *RoadMapBackground;
static GdkGC        *RoadMapSquareEdge;
static GdkGC        *RoadMapHighlight;
static RoadMapSprite RoadMapCursor;
static RoadMapSprite RoadMapDestination;
static RoadMapSprite RoadMapLocation;

static GtkWidget *RoadMapDrawingArea;
static GdkPixmap *RoadMapDrawingBuffer;


static struct roadmap_canvas_category {

   GdkGC *pen;
   int    declutter;

} RoadMapCategory [ROADMAP_CATEGORY_RANGE+1];


typedef struct roadmap_canvas_style {
   char *color;
   int thickness;
} RoadMapCanvasStyle;

static RoadMapCanvasStyle RoadMapBackgroundStyle;
static RoadMapCanvasStyle RoadMapPositionStyle;
static RoadMapCanvasStyle RoadMapSquareEdgeStyle;
static RoadMapCanvasStyle RoadMapHighlightStyle;
static RoadMapCanvasStyle RoadMapCategoryStyle [ROADMAP_CATEGORY_RANGE+1];


static char *SquareOnCanvas;
static int   SquareOnCanvasCount;


static void roadmap_canvas_refresh (void) {

   GdkRectangle update;

   update.x = 0;
   update.y = 0;
   update.width  = RoadMapDrawingArea->allocation.width;
   update.height = RoadMapDrawingArea->allocation.height;

   gtk_widget_draw (RoadMapDrawingArea, &update);
}


static void roadmap_canvas_set_thickness
               (GdkGC *gc, RoadMapCanvasStyle *style) {

   gdk_gc_set_line_attributes
      (gc,
       roadmap_math_thickness (style->thickness),
       GDK_LINE_SOLID, GDK_CAP_ROUND, GDK_JOIN_ROUND);
}


static GdkGC *roadmap_canvas_pen_new (RoadMapCanvasStyle *style) {

   GdkGC    *gc;
   GdkColor *color;

   color = (GdkColor *) g_malloc (sizeof(GdkColor));

   gdk_color_parse (style->color, color);
   gdk_color_alloc (gdk_colormap_get_system(), color);

   if (style->thickness <= 0) {
      style->thickness = 1;
   }

   gc = gdk_gc_new (RoadMapDrawingBuffer);

   gdk_gc_set_fill (gc, GDK_SOLID);

   gdk_gc_set_foreground (gc, color);

   roadmap_canvas_set_thickness (gc, style);

   return gc;
}


static void roadmap_canvas_draw_sprite (RoadMapSprite *sprite) {

   GdkPoint point;

   roadmap_math_coordinate (&(sprite->position), &point);

   gdk_draw_arc (RoadMapDrawingBuffer,
                 sprite->pen,
                 TRUE,
                 point.x - (ROADMAP_CURSOR_SIZE / 2),
                 point.y - (ROADMAP_CURSOR_SIZE / 2),
                 ROADMAP_CURSOR_SIZE,
                 ROADMAP_CURSOR_SIZE,
                 0,
                 (360 * 64));
}


static void roadmap_canvas_draw_information
               (RoadMapPosition *position, char *text, GdkGC *pen) {

   int y;
   GdkPoint points[7];

   int width  = RoadMapDrawingArea->allocation.width;
   int height = RoadMapDrawingArea->allocation.height;


   roadmap_math_coordinate (position, points);

   if (points[0].x < width / 2) {

      points[1].x = 10;
      points[2].x = 5;
      points[3].x = 5;
      points[4].x = width - 5;
      points[5].x = width - 5;
      points[6].x = 20;

   } else {

      points[1].x = width - 10;
      points[2].x = width - 5;
      points[3].x = width - 5;
      points[4].x = 5;
      points[5].x = 5;
      points[6].x = width - 20;
   }

   if (points[0].y < height / 2) {

      points[1].y = height - 25;
      points[2].y = height - 25;
      points[3].y = height - 5;
      points[4].y = height - 5;
      points[5].y = height - 25;
      points[6].y = height - 25;

      y = height - 9;

   } else {

      points[1].y = 25;
      points[2].y = 25;
      points[3].y = 5;
      points[4].y = 5;
      points[5].y = 25;
      points[6].y = 25;

      y = 21;
   }

   gdk_draw_polygon (RoadMapDrawingBuffer, pen, 1, points, 7);
   gdk_draw_polygon (RoadMapDrawingBuffer,
                     RoadMapDrawingArea->style->black_gc, 0, points, 7);

   gdk_draw_string  (RoadMapDrawingBuffer,
                     RoadMapDrawingArea->style->font,
                     RoadMapDrawingArea->style->black_gc, 9, y, text);
}


static void roadmap_canvas_draw_line_no_shape
               (int line, GdkGC *pen, int fully_visible) {

   RoadMapPosition position0;
   RoadMapPosition position1;

   GdkPoint points[2];


   roadmap_line_from (line, &position0);

   roadmap_line_to   (line, &position1);

   /* Optimization: do not draw lines that are obviously not visible. */

   if (! fully_visible)
   {
      int west;
      int east;
      int north;
      int south;

      if (position0.longitude < position1.longitude) {
         west = position0.longitude;
         east = position1.longitude;
      } else {
         west = position1.longitude;
         east = position0.longitude;
      }

      if (position0.latitude < position1.latitude) {
         south = position0.latitude;
         north = position1.latitude;
      } else {
         south = position1.latitude;
         north = position0.latitude;
      }

      if (! roadmap_math_is_visible (west, east, north, south)) return;
   }

   roadmap_math_coordinate (&position0, points);
   roadmap_math_coordinate (&position1, points+1);

   gdk_draw_lines (RoadMapDrawingBuffer, pen, points, 2);
}


static void roadmap_canvas_draw_line_with_shape
               (int line, int first_shape, int last_shape, GdkGC *pen) {

   int i;
   RoadMapPosition position_end;
   RoadMapPosition position;

   GdkPoint points[ROADMAP_CANVAS_POINT_BLOCK];

   int point_index;


   roadmap_line_to (line, &position_end);

   roadmap_line_from (line, &position);


   roadmap_math_coordinate (&position, points);
   point_index = 1;

   if (last_shape - first_shape + 3 >= ROADMAP_CANVAS_POINT_BLOCK) {

      roadmap_log (ROADMAP_ERROR, "cannot show all shape points.");
      last_shape = first_shape + ROADMAP_CANVAS_POINT_BLOCK - 3;
   }

   for (i = first_shape; i <= last_shape; i++) {

      roadmap_shape_get_position (i, &position);
      roadmap_math_coordinate (&position, points+point_index);

      point_index += 1;
   }

   roadmap_math_coordinate (&position_end, points+point_index);

   gdk_draw_lines (RoadMapDrawingBuffer, pen, points, point_index+1);
}


static void roadmap_canvas_draw_line (int line, GdkGC *pen) {

   int square;
   int first_shape_line;
   int last_shape_line;
   int first_shape;
   int last_shape;

   RoadMapPosition position;


   roadmap_line_from (line, &position);
   square = roadmap_square_search (&position);

   if (roadmap_shape_in_square (square, &first_shape_line,
                                        &last_shape_line) > 0) {

      if (roadmap_shape_of_line (line, first_shape_line,
                                       last_shape_line,
                                       &first_shape, &last_shape) > 0) {

         roadmap_canvas_draw_line_with_shape
               (line, first_shape, last_shape, pen);
         return;
      }
   }
   roadmap_canvas_draw_line_no_shape (line, pen, 1);
}


static void roadmap_canvas_draw_polygons (void) {

   int i;
   int j;
   int west;
   int east;
   int north;
   int south;
   int count;

   RoadMapPosition position;

   int points[ROADMAP_CANVAS_POINT_BLOCK];
   int size;

   GdkPoint graphic_points[ROADMAP_CANVAS_POINT_BLOCK];


   if (! roadmap_is_visible (ROADMAP_SHOW_AREA)) return;

   if (! roadmap_math_declutter (ROADMAP_DECLUTTER_POLYGONS)) return;

   count = roadmap_polygon_count ();

   for (i = 0; i < count; i++) {

      roadmap_polygon_edges (i, &west, &east, &north, &south);

      if (! roadmap_math_is_visible (west, east, north, south)) {
         continue;
      }

      size = roadmap_polygon_points (i, points, ROADMAP_CANVAS_POINT_BLOCK);

      for (j = 0; j < size; j++) {

         roadmap_point_position (points[j], &position);
         roadmap_math_coordinate (&position, graphic_points+j);
      }

      gdk_draw_polygon (RoadMapDrawingBuffer,
                        RoadMapCategory[roadmap_polygon_category(i)].pen, 1,
                        graphic_points, size);
   }
}


static void roadmap_canvas_draw_square_edges (int square) {

   RoadMapPosition topleft;
   RoadMapPosition bottomright;
   RoadMapPosition position;

   GdkPoint points[4];

   if (! roadmap_is_visible (ROADMAP_SHOW_SQUARE)) return;

   roadmap_square_edges (square, &(topleft.longitude),
                                 &(bottomright.longitude),
                                 &(topleft.latitude),
                                 &(bottomright.latitude));

   roadmap_math_coordinate (&topleft, points);

   position.longitude = bottomright.longitude;
   position.latitude  = topleft.latitude;
   roadmap_math_coordinate (&position, points+1);

   roadmap_math_coordinate (&bottomright, points+2);

   position.longitude = topleft.longitude;
   position.latitude  = bottomright.latitude;
   roadmap_math_coordinate (&position, points+3);

   gdk_draw_lines (RoadMapDrawingBuffer, RoadMapSquareEdge, points, 4);
}


static int roadmap_canvas_draw_square
              (int square, int cfcc, int fully_visible) {

   int line;
   int first_line;
   int last_line;
   int first_shape_line;
   int last_shape_line;
   int first_shape;
   int last_shape;

   int shapes_are_visible;

   int drawn = 0;


   GdkGC *pen = RoadMapCategory[cfcc].pen;


   roadmap_canvas_draw_square_edges (square);

   shapes_are_visible = roadmap_math_declutter (ROADMAP_DECLUTTER_SHAPE);


   /* Draw each line that belongs to this square. */

   if (roadmap_line_in_square (square, cfcc, &first_line, &last_line) > 0) {

      if (shapes_are_visible &&
          (roadmap_shape_in_square (square, &first_shape_line,
                                            &last_shape_line) > 0)) {

         for (line = first_line; line <= last_line; line++) {

            if (roadmap_shape_of_line (line, first_shape_line,
                                             last_shape_line,
                                             &first_shape, &last_shape) > 0) {
               roadmap_canvas_draw_line_with_shape
                  (line, first_shape, last_shape, pen);
            } else {
               roadmap_canvas_draw_line_no_shape (line, pen, fully_visible);
            }
            drawn += 1;
         }

      } else {

         for (line = first_line; line <= last_line; line++) {

            roadmap_canvas_draw_line_no_shape (line, pen, fully_visible);
            drawn += 1;
         }
      }
   }

   /* Draw each line that intersects with this square (but belongs
    * to another square--the crossing lines).
    */
   if (roadmap_line_in_square2 (square, cfcc, &first_line, &last_line) > 0) {

      int last_real_square = -1;
      int real_square  = 0;
      int check_shapes = 0;
      int west = 0;
      int east = 0;
      int north = 0;
      int south = 0;

      int real_line;
      int square_count = roadmap_square_count();
      char *on_canvas  = calloc (square_count, sizeof(char));

      for (line = first_line; line <= last_line; line++) {

         RoadMapPosition position;


         real_line = roadmap_line_get_from_index2 (line);

         roadmap_line_from (real_line, &position);

         /* Optimization: search for the square only if the new line does not
          * belong to the same square as the line before.
          */
         if ((position.longitude <  west) ||
             (position.longitude >= east) ||
             (position.latitude <  south) ||
             (position.latitude >= north)) {
            real_square = roadmap_square_search (&position);
         }

         if (real_square != last_real_square) {

            if (real_square < 0 || real_square >= square_count) {
               roadmap_log (ROADMAP_ERROR,
                            "Invalid square index %d", real_square);
               continue;
            }

            roadmap_square_edges (real_square, &west, &east, &north, &south);

            last_real_square = real_square;

            if (on_canvas[real_square]) {
               /* Either it has already been drawn, or it will be soon. */
               continue;
            }

            if (roadmap_math_is_visible (west, east, north, south)) {

               on_canvas[real_square] = 1;
               continue;
            }

            if (shapes_are_visible) {
               check_shapes = roadmap_shape_in_square
                                 (real_square,
                                  &first_shape_line, &last_shape_line);
            } else {
               check_shapes = 0;
            }
         }

         if (on_canvas[real_square]) {
            /* Either it has already been drawn, or it will be soon. */
            continue;
         }

         if (check_shapes) {
            if (roadmap_shape_of_line (real_line,
                                       first_shape_line,
                                       last_shape_line,
                                       &first_shape, &last_shape) > 0) {
               roadmap_canvas_draw_line_with_shape
                  (real_line, first_shape, last_shape, pen);
            } else {
               roadmap_canvas_draw_line_no_shape
                          (real_line, pen, fully_visible);
            }
         } else {
            roadmap_canvas_draw_line_no_shape (real_line, pen, fully_visible);
         }
         drawn += 1;
      }

      free (on_canvas);
   }

   return drawn;
}


static int roadmap_canvas_redraw (int square, int fully_visible) {

   int i;
   int drawn = 0;

   for (i = ROADMAP_CATEGORY_RANGE; i >= 1; i--) {

      if (roadmap_math_declutter (RoadMapCategory[i].declutter)) {

         roadmap_canvas_set_thickness
            (RoadMapCategory[i].pen, &(RoadMapCategoryStyle[i]));

         drawn += roadmap_canvas_draw_square (square, i, fully_visible);
      }
   }

   return drawn;
}


static int roadmap_canvas_draw_recursive (int square) {

   int i;
   int west;
   int east;
   int north;
   int south;
   int neighbor_count;
   int neighbors[8];

   int fully_visible;
   int drawn;


   if (SquareOnCanvas[square]) {
      return 0;
   }
   SquareOnCanvas[square] = 1;

   roadmap_square_edges (square, &west, &east, &north, &south);

   switch (roadmap_math_is_visible (west, east, north, south)) {
   case 0:
      return 0;
   case 1:
      fully_visible = 1;
      break;
   default:
      fully_visible = 0;
   }

   drawn = roadmap_canvas_redraw (square, fully_visible);

   neighbor_count = roadmap_square_neighborhood (square, neighbors);

   for (i = 0; i < neighbor_count; i++) {
      drawn += roadmap_canvas_draw_recursive (neighbors[i]);
   }
   return drawn;
}


static void roadmap_canvas_reset_square_mask (void) {

   if (SquareOnCanvas != NULL) {
      free(SquareOnCanvas);
   }
   SquareOnCanvasCount = roadmap_square_count();
   SquareOnCanvas = calloc (SquareOnCanvasCount, sizeof(char));
   if (SquareOnCanvas == NULL) {
      roadmap_log (ROADMAP_FATAL, "no more memory");
   }
}


static int roadmap_canvas_repaint_square (int square) {

   roadmap_canvas_reset_square_mask();

   return roadmap_canvas_draw_recursive (square);
}


static void roadmap_canvas_repaint (void) {

   static int *fips = NULL;

   int i;
   int j;
   int count;
   int square;


   /* Erase the drawing buffer. */

   gdk_draw_rectangle (RoadMapDrawingBuffer,
                       RoadMapBackground,
                       TRUE,
                       0, 0,
                       RoadMapDrawingArea->allocation.width,
                       RoadMapDrawingArea->allocation.height);


   /* Repaint the drawing buffer. */

   /* - Identifies the candidate counties. */

   count = roadmap_locator_by_position (&(RoadMapCursor.position), &fips);


   /* - For each candidate county: */

   for (i = 0; i < count; i++) {

      /* -- Access the county's database. */

      if (roadmap_locator_activate (fips[i]) != ROADMAP_US_OK) continue;

      /* -- Look for the square the current location fits in. */

      square = roadmap_square_search (&(RoadMapCursor.position));

      if (square < 0) {

         /* The current location does not fit in any particular square:
          * we are not in that county. Is any portion of this county's
          * squares is visible ?
          */
         int closest;
         int neighbors[1024];

         closest = roadmap_square_closest
                      (&(RoadMapCursor.position), neighbors, 1024);

         if (closest > 0) {

            roadmap_canvas_draw_polygons ();
            roadmap_canvas_reset_square_mask();

            for (j = 0; j < closest; j++) {
               roadmap_canvas_draw_recursive (neighbors[j]);
            }
         }

      } else {

         /* The current location fits in one of the county's squares.
          * We might be in that county, and this square is visible, anyway.
          */
         roadmap_canvas_draw_polygons ();
         roadmap_canvas_repaint_square (square);
      }
   }

   roadmap_canvas_draw_sprite (&RoadMapLocation);
   roadmap_canvas_draw_sprite (&RoadMapDestination);
}


static gint roadmap_canvas_configure
               (GtkWidget *widget, GdkEventConfigure *event) {

   int i;

   if (RoadMapDrawingBuffer != NULL) {

      gdk_pixmap_unref (RoadMapDrawingBuffer);
   }

   RoadMapDrawingBuffer =
      gdk_pixmap_new (widget->window,
                      widget->allocation.width,
                      widget->allocation.height,
                      -1);

   for (i = 1; i <= ROADMAP_CATEGORY_RANGE; i++) {
      if (RoadMapCategory[i].pen == NULL) {
         RoadMapCategory[i].pen =
            roadmap_canvas_pen_new (&(RoadMapCategoryStyle[i]));
      }
   }

   if (RoadMapBackground == NULL) {
      RoadMapBackground = roadmap_canvas_pen_new (&RoadMapBackgroundStyle);
   }

   if (RoadMapHighlight == NULL) {
      RoadMapHighlight = roadmap_canvas_pen_new (&RoadMapHighlightStyle);
   }

   if (roadmap_is_visible (ROADMAP_SHOW_SQUARE)) {
      if (RoadMapSquareEdge == NULL) {
         RoadMapSquareEdge = roadmap_canvas_pen_new (&RoadMapSquareEdgeStyle);
      }
   }

   if (RoadMapDestination.pen == NULL) {
      RoadMapPositionStyle.color = 
         roadmap_config_get ("Point Colors", "Endpoint");
      RoadMapDestination.pen = roadmap_canvas_pen_new (&RoadMapPositionStyle);
   }
   if (RoadMapLocation.pen == NULL) {
      RoadMapPositionStyle.color = 
         roadmap_config_get ("Point Colors", "Position");
      RoadMapLocation.pen = roadmap_canvas_pen_new (&RoadMapPositionStyle);
   }

   roadmap_math_set_size (widget->allocation.width, widget->allocation.height);

   if ((RoadMapCursor.position.latitude != 0) &&
       (RoadMapCursor.position.longitude != 0)) {

      roadmap_canvas_repaint ();
      roadmap_canvas_refresh ();
   }

   return TRUE;
}


static gint roadmap_canvas_expose (GtkWidget *widget, GdkEventExpose *event) {

   gdk_draw_pixmap (widget->window,
                    widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
                    RoadMapDrawingBuffer,
                    event->area.x, event->area.y,
                    event->area.x, event->area.y,
                    event->area.width, event->area.height);

   return FALSE;
}


static gint roadmap_canvas_button_pressed
               (GtkWidget *w, GdkEventButton *event, gpointer data) {

   int i;
   int detail = 0;
   GdkPoint point;
   RoadMapPosition position;

   int line;
   int distance;

   char *name;


   point.x = event->x;
   point.y = event->y;
   roadmap_math_to_position (&point, &position);
#ifdef DEBUG
printf ("Position: %d longitude, %d latitude\n",
        position.longitude,
        position.latitude);
fflush(stdout);
#endif

   for (i = ROADMAP_ROAD_FIRST; i <= ROADMAP_ROAD_LAST; i++) {

      if (roadmap_math_declutter (RoadMapCategory[i].declutter)) {

         detail = i;
      }
   }

   if (detail > 0) {

      roadmap_canvas_repaint ();

      line = roadmap_street_get_closest (&position, detail, &distance);

      if (line >= 0) {

         name = roadmap_street_get_name_from_line (line);

         roadmap_canvas_draw_line (line, RoadMapHighlight);

         roadmap_canvas_draw_information (&position, name, RoadMapHighlight);
      }
      roadmap_canvas_refresh ();
   }

   return FALSE;
}


static void roadmap_canvas_draw (RoadMapPosition *position) {

   RoadMapCursor.position = *position;

   roadmap_math_set_center (&(RoadMapCursor.position));

   roadmap_canvas_repaint ();
   roadmap_canvas_refresh ();
}


void roadmap_canvas_initialize (void) {

   roadmap_config_declare_preference ("General", "Background", "LightYellow");

   roadmap_config_declare_preference ("Highlight", "Color", "yellow");
   roadmap_config_declare_preference ("Highlight", "Thickness", "4");

   roadmap_config_declare_preference ("Point Colors", "Endpoint", "green");
   roadmap_config_declare_preference ("Point Colors", "Waypoint", "yellow");
   roadmap_config_declare_preference ("Point Colors", "Position", "BlueViolet");

   roadmap_config_declare_preference ("Line Colors", "Freeways",  "red");
   roadmap_config_declare_preference ("Line Colors", "Shore",     "SlateBlue");
   roadmap_config_declare_preference ("Line Colors", "Rivers",    "LightBlue");
   roadmap_config_declare_preference ("Line Colors", "Lakes",     "LightBlue");

   roadmap_config_declare_preference ("Block Colors", "Parks",     "green");
   roadmap_config_declare_preference ("Block Colors", "Hospitals", "red");
   roadmap_config_declare_preference ("Block Colors", "Airports",  "grey");
   roadmap_config_declare_preference ("Block Colors", "Stations",  "grey");
   roadmap_config_declare_preference ("Block Colors", "Malls",     "yellow");

   roadmap_config_declare_session ("Locations", "Position", "0,0");
   roadmap_config_declare_session ("Locations", "Destination", "0,0");
}


GtkWidget *roadmap_canvas (void) {

   int i;


   RoadMapDrawingArea = gtk_drawing_area_new ();

   gtk_widget_set_events (RoadMapDrawingArea, GDK_BUTTON_PRESS_MASK);

   gtk_signal_connect (GTK_OBJECT(RoadMapDrawingArea),
                       "expose_event",
                       (GtkSignalFunc) roadmap_canvas_expose,
                       NULL);

   gtk_signal_connect (GTK_OBJECT(RoadMapDrawingArea),
                       "configure_event",
                       (GtkSignalFunc) roadmap_canvas_configure,
                       NULL);

   gtk_signal_connect (GTK_OBJECT(RoadMapDrawingArea),
                       "button_press_event",
                       GTK_SIGNAL_FUNC(roadmap_canvas_button_pressed),
                       NULL);


   for (i = 1; i <= ROADMAP_CATEGORY_RANGE; i++) {

      RoadMapCategory[i].declutter = 0;
      RoadMapCategoryStyle[i].thickness = 4;
      RoadMapCategoryStyle[i].color = "grey10";
   }

   RoadMapBackgroundStyle.color =
                  roadmap_config_get ("General", "Background");

   RoadMapSquareEdgeStyle.color = "grey";
   RoadMapSquareEdgeStyle.thickness = 4;

   RoadMapHighlightStyle.color =
                  roadmap_config_get ("Highlight", "Color");
   RoadMapHighlightStyle.thickness =
                  roadmap_config_get_integer ("Highlight", "Thickness");

   RoadMapCategory[ROADMAP_ROAD_MAIN].declutter    = ROADMAP_DECLUTTER_MAIN;
   RoadMapCategory[ROADMAP_ROAD_STREET].declutter  = ROADMAP_DECLUTTER_STREET;
   RoadMapCategory[ROADMAP_ROAD_FREEWAY].declutter = ROADMAP_DECLUTTER_FREEWAY;
   RoadMapCategory[ROADMAP_WATER_SHORELINE].declutter = ROADMAP_DECLUTTER_SEA;
   RoadMapCategory[ROADMAP_WATER_RIVER].declutter = ROADMAP_DECLUTTER_RIVER;
   RoadMapCategory[ROADMAP_WATER_LAKE].declutter = ROADMAP_DECLUTTER_LAKE;
   RoadMapCategory[ROADMAP_WATER_SEA].declutter = ROADMAP_DECLUTTER_SEA;

   RoadMapCategoryStyle[ROADMAP_ROAD_STREET].thickness = ROADMAP_THICKNESS_STREET;
   RoadMapCategoryStyle[ROADMAP_ROAD_MAIN].thickness   = ROADMAP_THICKNESS_MAIN;
   RoadMapCategoryStyle[ROADMAP_ROAD_FREEWAY].thickness = ROADMAP_THICKNESS_FREEWAY;
   RoadMapCategoryStyle[ROADMAP_WATER_SHORELINE].thickness = ROADMAP_THICKNESS_SHORE;
   RoadMapCategoryStyle[ROADMAP_WATER_RIVER].thickness = ROADMAP_THICKNESS_RIVER;
   RoadMapCategoryStyle[ROADMAP_WATER_LAKE].thickness = ROADMAP_THICKNESS_RIVER;

   RoadMapCategoryStyle[ROADMAP_ROAD_FREEWAY].color =
                  roadmap_config_get ("Line Colors", "Freeways");
   RoadMapCategoryStyle[ROADMAP_WATER_SHORELINE].color =
                  roadmap_config_get ("Line Colors", "Shore");
   RoadMapCategoryStyle[ROADMAP_WATER_RIVER].color =
                  roadmap_config_get ("Line Colors", "Rivers");

   RoadMapCategoryStyle[ROADMAP_WATER_LAKE].color =
                  roadmap_config_get ("Block Colors", "Lakes");

   RoadMapCategoryStyle[ROADMAP_AREA_PARC].color =
                  roadmap_config_get ("Block Colors", "Parks");
   RoadMapCategoryStyle[ROADMAP_AREA_HOSPITAL].color =
                  roadmap_config_get ("Block Colors", "Hospitals");
   RoadMapCategoryStyle[ROADMAP_AREA_AIRPORT].color =
                  roadmap_config_get ("Block Colors", "Airports");
   RoadMapCategoryStyle[ROADMAP_AREA_STATION].color =
                  roadmap_config_get ("Block Colors", "Stations");
   RoadMapCategoryStyle[ROADMAP_AREA_MALL].color =
                  roadmap_config_get ("Block Colors", "Malls");


   return RoadMapDrawingArea;
}

void roadmap_canvas_set_initial_position (void) {

   roadmap_config_get_position ("Position", &RoadMapLocation.position);
   roadmap_config_get_position ("Destination", &RoadMapDestination.position);

   roadmap_canvas_draw (&RoadMapLocation.position);
}


void roadmap_canvas_set_location (RoadMapPosition *position) {

   if (position != NULL) {

      if ((position->longitude != RoadMapLocation.position.longitude)
             || (position->latitude != RoadMapLocation.position.latitude)) {

         roadmap_config_set_position ("Position", position);
         RoadMapLocation.position = *position;
      }
   }

   roadmap_canvas_draw (&RoadMapLocation.position);
}


void roadmap_canvas_set_destination (RoadMapPosition *position) {

   if (position != NULL) {

      if ((position->longitude != RoadMapDestination.position.longitude)
             || (position->latitude != RoadMapDestination.position.latitude)) {

         roadmap_config_set_position ("Destination", position);
         RoadMapDestination.position = *position;
      }
   }

   roadmap_canvas_draw (&RoadMapDestination.position);
}


void roadmap_canvas_move_north (void) {

   int north;
   int south;

   roadmap_math_screen_edges (NULL, NULL, &north, &south);

   RoadMapCursor.position.latitude += abs(north - south) / 4;
   roadmap_canvas_draw (&RoadMapCursor.position);
}


void roadmap_canvas_move_south (void) {

   int north;
   int south;

   roadmap_math_screen_edges (NULL, NULL, &north, &south);

   RoadMapCursor.position.latitude -= abs(north - south) / 4;
   roadmap_canvas_draw (&RoadMapCursor.position);
}


void roadmap_canvas_move_east (void) {

   int west;
   int east;

   roadmap_math_screen_edges (&west, &east, NULL, NULL);

   RoadMapCursor.position.longitude += abs(west - east) / 4;
   roadmap_canvas_draw (&RoadMapCursor.position);
}


void roadmap_canvas_move_west (void) {

   int west;
   int east;

   roadmap_math_screen_edges (&west, &east, NULL, NULL);

   RoadMapCursor.position.longitude -= abs(west - east) / 4;
   roadmap_canvas_draw (&RoadMapCursor.position);
}


void roadmap_canvas_zoom_in  (void) {

   roadmap_math_zoom_in   ();
   roadmap_canvas_repaint ();
   roadmap_canvas_refresh ();
}


void roadmap_canvas_zoom_out (void) {

   roadmap_math_zoom_out  ();
   roadmap_canvas_repaint ();
   roadmap_canvas_refresh ();
}


void roadmap_canvas_zoom_reset (void) {

   roadmap_math_zoom_reset();
   roadmap_canvas_repaint ();
   roadmap_canvas_refresh ();
}


void roadmap_canvas_close (void) {
}
