/* roadmap_preferences.c - handle the roadmap dialogs managing user preferences.
 *
 * LICENSE:
 *
 *   Copyright 2002 Pascal F. Martin
 *
 *   This file is part of RoadMap.
 *
 *   RoadMap is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   RoadMap is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with RoadMap; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * SYNOPSYS:
 *
 *   void roadmap_preferences_edit (void);
 */

#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "roadmap.h"
#include "roadmap_types.h"
#include "roadmap_gui.h"
#include "roadmap_preferences.h"


typedef struct EditableItemRecord {

   struct EditableItemRecord *next;

   char *category;
   char *name;

   GtkWidget *entry;
   GtkWidget *dialog;

} EditableItem ;

static gint roadmap_preferences_cancel (GtkWidget *w, gpointer data) {

   EditableItem *next;
   EditableItem *item = (EditableItem *) data;
   GtkWidget *dialog;

   if (item != NULL) {

      dialog = item->dialog;

      for (item = (EditableItem *) data; item != NULL; item = next) {
         next = item->next;
         free (item);
      }

      gtk_grab_remove (dialog);
      gtk_widget_destroy (dialog);
   }

   return FALSE;
}

static gint roadmap_preferences_ok (GtkWidget *w, gpointer data) {

   EditableItem *item;

   for (item = (EditableItem *) data; item != NULL; item = item->next) {

      roadmap_config_set
         (item->category,
          item->name,
          gtk_entry_get_text (GTK_ENTRY(item->entry)));
   }

   return roadmap_preferences_cancel (w, data);
}

static void roadmap_preferences_new (char *title, void *cursor) {

   char full_title[256];

   int i;
   int count = 0;
   int lines[256];
   char *categories[256];
   GtkWidget *pages[256];

   char *category;
   char *name;
   char *value;

   GtkWidget *ok;
   GtkWidget *cancel;
   GtkWidget *label;
   GtkWidget *entry;
   GtkWidget *dialog;
   GtkWidget *notebook;

   EditableItem *new_item;
   EditableItem *last_item = NULL;


   if (cursor == NULL) return;

   dialog = gtk_dialog_new();

   roadmap_window_set_size (dialog, 400, 300);

   strcpy (full_title, "RoadMap ");
   strcat (full_title, title);
   gtk_window_set_title (GTK_WINDOW(dialog), full_title);

   notebook = gtk_notebook_new();

   gtk_notebook_set_scrollable (GTK_NOTEBOOK(notebook), TRUE);

   gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dialog)->vbox),
                       notebook, TRUE, TRUE, 0);

   while (cursor != NULL) {

      cursor = roadmap_config_scan (cursor, &category, &name, &value);

      for (i = count - 1; i >= 0; i--) {
         if (strcmp (category, categories[i]) == 0) break;
      }

      if (i < 0) {

         /* This is a new category. */

         if (count >= 256) continue; /* Cannot handle more category. */

         label = gtk_label_new (category);

         lines[count] = 0;
         pages[count] = gtk_table_new (1, 2, FALSE);
         categories[count] = category;

         i = count;
         count += 1;

         gtk_notebook_append_page
            (GTK_NOTEBOOK(notebook), pages[i], label);

      } else {
         gtk_table_resize (GTK_TABLE(pages[i]), lines[i]+2, 2);
      }

      label = gtk_label_new (name);
      gtk_label_set_justify (GTK_LABEL(label), GTK_JUSTIFY_LEFT);
      gtk_label_set_text    (GTK_LABEL(label), name); // ?? does not justify ?

      gtk_table_attach (GTK_TABLE(pages[i]),
                        label, 0, 1, lines[i], lines[i]+1,
                        GTK_FILL, GTK_FILL, 2, 2);

      entry = gtk_entry_new ();
      gtk_entry_set_text (GTK_ENTRY(entry), value);

      gtk_table_attach (GTK_TABLE(pages[i]),
                        entry, 1, 2, lines[i], lines[i]+1,
                        GTK_EXPAND+GTK_FILL+GTK_SHRINK,
                        GTK_EXPAND+GTK_FILL+GTK_SHRINK, 2, 2);


      new_item = malloc (sizeof(EditableItem));
      if (new_item == NULL) {
         roadmap_log (ROADMAP_FATAL, "no more memory");
      }

      new_item->category = category;
      new_item->name   = name;
      new_item->entry  = entry;
      new_item->dialog = dialog;

      new_item->next = last_item;
      last_item = new_item;

      lines[i] += 1;
   }

   if (last_item == NULL) {
      roadmap_log (ROADMAP_ERROR, "no preference item registered");
   }

   gtk_container_set_border_width
      (GTK_CONTAINER(GTK_BOX(GTK_DIALOG(dialog)->vbox)), 4);


   ok     = gtk_button_new_with_label ("Ok");
   cancel = gtk_button_new_with_label ("Cancel");

   GTK_WIDGET_SET_FLAGS (ok, GTK_CAN_DEFAULT);
   GTK_WIDGET_SET_FLAGS (cancel, GTK_CAN_DEFAULT);

   gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dialog)->action_area),
                       ok, TRUE, TRUE, 0);

   gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dialog)->action_area),
                       cancel, TRUE, TRUE, 0);

   gtk_signal_connect (GTK_OBJECT(ok),
                       "clicked",
                       GTK_SIGNAL_FUNC(roadmap_preferences_ok),
                       last_item);

   gtk_signal_connect (GTK_OBJECT(cancel),
                       "clicked",
                       GTK_SIGNAL_FUNC(roadmap_preferences_cancel),
                       last_item);

   gtk_widget_grab_default (ok);

   gtk_grab_add (dialog);

   gtk_widget_show_all (GTK_WIDGET(dialog));
}


void roadmap_preferences_edit (void) {
   roadmap_preferences_new ("Preferences", roadmap_config_first_preference());
}

