/* roadmap_config.c - A module to handle all RoadMap configuration issues.
 *
 * LICENSE:
 *
 *   Copyright 2002 Pascal F. Martin
 *
 *   This file is part of RoadMap.
 *
 *   RoadMap is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   RoadMap is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with RoadMap; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * SYNOPSYS:
 *
 *   See roadmap_config.h.
 */

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "roadmap_dbread.h"

#include "roadmap.h"
#include "roadmap_math.h"
#include "roadmap_config.h"


typedef struct RoadMapConfigEnumRecord RoadMapConfigEnum;

struct RoadMapConfigEnumRecord {

   RoadMapConfigEnum *next;

   char *value;
};


typedef struct RoadMapConfigItemRecord RoadMapConfigItem;

struct RoadMapConfigItemRecord {

   RoadMapConfigItem *next;

   char *category;
   char *name;
   char *default_value;
   char *value;

   int type;

   union {
      RoadMapConfigEnum *enumeration_values;
   } detail;
};


typedef struct {

   int                changed;
   char              *name;
   RoadMapConfigItem *items;

} RoadMapConfig;


static RoadMapConfig RoadMapSession     = {0, "session", NULL};
static RoadMapConfig RoadMapPreferences = {0, "preferences", NULL};


static RoadMapConfigItem *roadmap_config_declare
                             (RoadMapConfig *config,
                              char *category, char *name, char *default_value) {

   RoadMapConfigItem *new_item = malloc (sizeof(RoadMapConfigItem));

   if (new_item == NULL) {
      roadmap_log (ROADMAP_FATAL, "no more memory");
   }
   new_item->name = name;
   new_item->category = category;
   new_item->default_value = default_value;

   new_item->value = NULL;

   new_item->next = config->items;
   config->items  = new_item;

   return new_item;
}


static void roadmap_config_add_enumeration_value
               (RoadMapConfigItem *item, char *value) {

   RoadMapConfigEnum *new_value;
   RoadMapConfigEnum *last = item->detail.enumeration_values;

   new_value = malloc (sizeof(RoadMapConfigEnum));
   if (new_value == NULL) {
      roadmap_log (ROADMAP_FATAL, "no more memory");
   }
   new_value->next = NULL;
   new_value->value = value;

   if (last == NULL) {

      item->detail.enumeration_values = new_value;

   } else {

      while (last->next != NULL) last = last->next;

      last->next = new_value;
   }
}


void roadmap_config_declare_preference
        (char *category, char *name, char *default_value) {

   RoadMapConfigItem *item;

   item = roadmap_config_declare
             (&RoadMapPreferences, category, name, default_value);
   item->type = ROADMAP_CONFIG_STRING;
}


void roadmap_config_declare_enumeration_preference
        (char *category, char *name, char *enumeration_value, ...) {

   char *p;
   va_list ap;
   RoadMapConfigItem *item;

   item = roadmap_config_declare
             (&RoadMapPreferences, category, name, enumeration_value);
   item->type = ROADMAP_CONFIG_ENUM;

   roadmap_config_add_enumeration_value (item, enumeration_value);

   va_start(ap, enumeration_value);
   for (p = va_arg(ap, char *); p != NULL; p = va_arg(ap, char *)) {
      roadmap_config_add_enumeration_value (item, p);
   }
}


void roadmap_config_declare_color_preference
        (char *category, char *name, char *default_value) {

   RoadMapConfigItem *item;

   item = roadmap_config_declare
             (&RoadMapPreferences, category, name, default_value);
   item->type = ROADMAP_CONFIG_COLOR;
}


void roadmap_config_declare_session
        (char *category, char *name, char *default_value) {

   roadmap_config_declare (&RoadMapSession, category, name, default_value);
}


void *roadmap_config_first_preference (void) {

   return RoadMapPreferences.items;
}

void *roadmap_config_first_session (void) {

   return RoadMapSession.items;
}

int   roadmap_config_get_type (void *cursor) {

   RoadMapConfigItem *item = (RoadMapConfigItem *)cursor;

   if (item == NULL) {
      return ROADMAP_CONFIG_STRING;
   }

   return item->type;
}


void *roadmap_config_scan
         (void *cursor, char **category, char **name, char **value) {

   RoadMapConfigItem *item = (RoadMapConfigItem *)cursor;

   if (item == NULL) {
      return NULL;
   }

   *category = item->category;
   *name     = item->name;
   *value    = item->value;

   return item->next;
}


void *roadmp_config_get_enumeration (void *cursor) {

   RoadMapConfigItem *item = (RoadMapConfigItem *)cursor;

   if (item == NULL) {
      return NULL;
   }

   if (item->type != ROADMAP_CONFIG_ENUM) {
      return NULL;
   }

   return item->detail.enumeration_values;
}


char *roadmp_config_get_enumeration_value (void *enumeration) {

   RoadMapConfigEnum *item = (RoadMapConfigEnum *)enumeration;

   if (item == NULL) { 
      return NULL;
   }

   return item->value;
}


void *roadmp_config_get_enumeration_next (void *enumeration) {

   RoadMapConfigEnum *item = (RoadMapConfigEnum *)enumeration;

   if (item == NULL) { 
      return NULL;
   }

   return item->next;
}


static int roadmap_config_set_item
               (RoadMapConfig *config,
                char *category, char *name, char *value) {

   RoadMapConfigItem *item;

   for (item = config->items; item != NULL; item = item->next) {

      if ((strcmp (item->name, name) == 0) &&
          (strcmp (item->category, category) == 0)) {

         if (item->value != NULL) {
            if (value != NULL) {
               if (strcmp (item->value, value) == 0) {
                  return 1; /* Nothing was changed. */
               }
            }
            free(item->value);
         }
         if (value != NULL) {
            item->value = strdup(value);
         } else {
            item->value = NULL;
         }

         config->changed = 1;
         return 1;
      }
   }
   return 0;
}


static char *roadmap_config_get_item
                (RoadMapConfig *config, char *category, char *name) {

   RoadMapConfigItem *item;

   for (item = config->items; item != NULL; item = item->next) {

      if ((strcmp (item->name, name) == 0) &&
          (strcmp (item->category, category) == 0)) {

         if (item->value != NULL) {
            return item->value;
         }
         return item->default_value;
      }
   }
   return "";
}


static char *roadmap_config_skip_until (char *p, char c) {

   while (*p != '\n' && *p != c && *p > 0) p++;
   return p;
}

static char *roadmap_config_skip_spaces (char *p) {

   while (*p == ' ' || *p == '\t') p++;
   return p;
}


static void roadmap_config_load (RoadMapConfig *config) {

   char *p;
   FILE *file;
   char  line[1024];

   char *category;
   char *name;
   char *value;

   RoadMapConfigItem *item;


   file = roadmap_config_open (config->name, "r");

   if (file == NULL) {
      config->changed = 1; /* Force the creation of the file on exit. */
      return;
   }

   while (!feof(file)) {

      if (fgets (line, sizeof(line), file) == NULL) break;

      p = roadmap_config_skip_spaces (line);

      if (*p == '\n' || *p == '#') continue;

      category = p;

      p = roadmap_config_skip_until (p, '.');
      if (*p != '.') continue;
      *(p++) = 0;

      name = p;

      p = roadmap_config_skip_until (p, ':');
      if (*p != ':') continue;
      *(p++) = 0;

      p = roadmap_config_skip_spaces (p);
      value = p;

      p = roadmap_config_skip_until (p, 0);
      *p = 0;

      roadmap_config_set_item (config, category, name, value);
   }

   config->changed = 0;
   fclose (file);

   for (item = config->items; item != NULL; item = item->next) {

      if (item->value == NULL) {

         config->changed = 1;
         break;
      }
   }
}


static void roadmap_config_save (RoadMapConfig *config) {

   FILE *file;
   char *value;
   RoadMapConfigItem *item;

   if (config->changed) {

      file = roadmap_config_open (config->name, "w");

      for (item = config->items; item != NULL; item = item->next) {

         if (item->value != NULL) {
            value = item->value;
         } else {
            value = item->default_value;
         }
         fprintf (file, "%s.%s: %s\n", item->category, item->name, value);
      }

      fclose (file);
      config->changed = 0;
   }
}


FILE *roadmap_config_open (char *name, char *mode) {

   FILE *file;
   char path[1024];

   strcpy (path, getenv("HOME"));
   strcat (path, "/.roadmap/");
   mkdir  (path, 0755);

   strcat (path, name);

   file = fopen (path, mode);

   if (file == NULL) {
      if (*mode == 'w') {
         roadmap_log (ROADMAP_FATAL, "cannot open file %s", path);
      } else {
         roadmap_log (ROADMAP_ERROR, "cannot open file %s", path);
      }
   }

   return file;
}


void  roadmap_config_initialize (void) {

   roadmap_config_declare_preference
      ("General", "Database", "~/maps,/usr/share/roadmap/maps");

   roadmap_config_load (&RoadMapPreferences);
   roadmap_config_load (&RoadMapSession);

   roadmap_db_path
      (roadmap_config_get_item (&RoadMapPreferences, "General", "Database"));
}


void roadmap_config_update (void) {

   roadmap_config_save (&RoadMapPreferences);
   roadmap_config_save (&RoadMapSession);
}


char *roadmap_config_get (char *category, char *name) {

   char *value;

   value = roadmap_config_get_item (&RoadMapSession, category, name);

   if (value[0] == 0) {
      value = roadmap_config_get_item (&RoadMapPreferences, category, name);
   }

   return value;
}


int roadmap_config_get_integer(char *category, char *name) {

   return atoi (roadmap_config_get (category, name));
}


void  roadmap_config_set (char *category, char *name, char *value) {

   if (! roadmap_config_set_item (&RoadMapSession, category, name, value)) {
      roadmap_config_set_item (&RoadMapPreferences, category, name, value);
   }
}


void roadmap_config_get_position (char *name, RoadMapPosition *position) {

   char *center = roadmap_config_get_item (&RoadMapSession, "Locations", name);

   if (center != NULL && center[0] != 0) {

      char *latitude_ascii;
      char buffer[128];


      strcpy (buffer, center);
      latitude_ascii = strchr (buffer, ',');
      if (latitude_ascii != NULL) {
         *(latitude_ascii++) = 0;
      }

      position->longitude = atoi(buffer);
      position->latitude  = atoi(latitude_ascii);

   } else {

      position->longitude = 0;
      position->latitude  = 0;
   }
}

void  roadmap_config_set_position (char *name, RoadMapPosition *position) {

   char buffer[128];

   sprintf (buffer, "%d,%d", position->longitude, position->latitude);
   roadmap_config_set_item (&RoadMapSession, "Locations", name, buffer);
}

