/* roadmap_dialog.c - manage the roadmap dialogs used for user input.
 *
 * LICENSE:
 *
 *   Copyright 2002 Pascal F. Martin
 *
 *   This file is part of RoadMap.
 *
 *   RoadMap is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   RoadMap is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with RoadMap; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * SYNOPSYS:
 *
 *   void roadmap_dialog_destination_address (void);
 *   void roadmap_dialog_location_address (void);
 *   void roadmap_dialog_about (void);
 */

#include <math.h>
#include <stdlib.h>

#include "roadmap.h"
#include "roadmap_types.h"
#include "roadmap_math.h"
#include "roadmap_gui.h"
#include "roadmap_street.h"
#include "roadmap_locator.h"
#include "roadmap_dialog.h"


typedef void (*roadmap_process_address) (RoadMapPosition *position);

typedef struct {

   char character;

   GtkWidget *button;
   void *context;

} RoadMapKey;


typedef struct {

   char *title;
   int   use_zip;
   roadmap_process_address when_done;

   GtkWidget *dialog;

   GtkWidget *entry[4];

   GtkWidget *selected;

   RoadMapKey keyboard[40];

} RoadMapAddressDialog;


static gint roadmap_dialog_message_ok (GtkWidget *w, gpointer data) {

   GtkWidget *dialog = (GtkWidget *) data;

   gtk_grab_remove (dialog);
   gtk_widget_destroy (dialog);

   return FALSE;
}

static void roadmap_dialog_message (char *text) {

   GtkWidget *ok;
   GtkWidget *label;
   GtkWidget *dialog;

   dialog = gtk_dialog_new();

   label = gtk_label_new(text);

   ok = gtk_button_new_with_label ("Ok");

   GTK_WIDGET_SET_FLAGS (ok, GTK_CAN_DEFAULT);

   gtk_window_set_title (GTK_WINDOW(dialog), "RoadMap");

   gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dialog)->vbox),
                       label, TRUE, TRUE, 0);

   gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dialog)->action_area),
                       ok, TRUE, TRUE, 0);

   gtk_container_set_border_width
      (GTK_CONTAINER(GTK_BOX(GTK_DIALOG(dialog)->vbox)), 4);

   gtk_signal_connect (GTK_OBJECT(ok),
                       "clicked",
                       GTK_SIGNAL_FUNC(roadmap_dialog_message_ok),
                       dialog);

   gtk_widget_grab_default (ok);

   gtk_grab_add (dialog);

   gtk_widget_show_all (GTK_WIDGET(dialog));
}


static gint roadmap_dialog_selected
               (GtkWidget *w, GdkEventButton *event, gpointer data) {

   int i;
   RoadMapAddressDialog *context = (RoadMapAddressDialog *) data;

   for (i = 0; i < 4; i++) {
      if (context->entry[i] == w) {
         context->selected = context->entry[i];
      }
   }

   return FALSE;
}


static gint roadmap_dialog_key_pressed (GtkWidget *w, gpointer data) {

   gint cursor;

   RoadMapKey *key = (RoadMapKey *) data;
   RoadMapAddressDialog *context = (RoadMapAddressDialog *) key->context;


   if (context->selected != NULL) {

      cursor = gtk_editable_get_position (GTK_EDITABLE(context->selected));

      if (key->character >= ' ') {

         char text[2];

         text[0] = key->character;
         text[1] = 0;
         gtk_editable_insert_text (GTK_EDITABLE(context->selected),
                                   text, 1, &cursor);
         gtk_editable_set_position (GTK_EDITABLE(context->selected), cursor);

      } else if (key->character == '\b') {

         cursor = gtk_editable_get_position (GTK_EDITABLE(context->selected));

         gtk_editable_delete_text
            (GTK_EDITABLE(context->selected), cursor-1, cursor);

      } else if (key->character == '\r') {

         gtk_editable_delete_text (GTK_EDITABLE(context->selected), 0, -1);
      }
      gtk_widget_grab_focus (context->selected);
   }

   return FALSE;
}


static gint roadmap_dialog_address_cancel (GtkWidget *w, gpointer data) {

   RoadMapAddressDialog *context = (RoadMapAddressDialog *) data;

   gtk_grab_remove (context->dialog);
   gtk_widget_hide (context->dialog);

   return FALSE;
}

static gint roadmap_dialog_address_ok (GtkWidget *w, gpointer data) {

   int count;
   RoadMapBlocks blocks[256];

   int   line;
   int   street_number;
   char *street_number_image;
   char *street_name;
   char *city;
   char *state;
   RoadMapPosition position;
   RoadMapAddressDialog *context = (RoadMapAddressDialog *) data;


   street_number_image = gtk_entry_get_text (GTK_ENTRY(context->entry[0]));
   street_number =
      roadmap_math_street_address
         (street_number_image, strlen(street_number_image));

   street_name   = gtk_entry_get_text (GTK_ENTRY(context->entry[1]));
   city          = gtk_entry_get_text (GTK_ENTRY(context->entry[2]));
   state         = gtk_entry_get_text (GTK_ENTRY(context->entry[3]));

   if (context->use_zip) {
      // TBD: how to select by ZIP ? Need one more table in usdir.
      // count = roadmap_street_blocks_by_zip
                 // (street_name, atoi(city), blocks, 256);
      return FALSE;
   }

   switch (roadmap_locator_by_city (city, state)) {

   case ROADMAP_US_OK:

      count = roadmap_street_blocks_by_city (street_name, city, blocks, 256);
      break;

   case ROADMAP_US_NOSTATE:

      roadmap_dialog_message ("No state with that name could be found");
      return FALSE;

   case ROADMAP_US_NOCITY:

      roadmap_dialog_message ("No city with that name could be found");
      return FALSE;

   default:

      roadmap_dialog_message ("No related map could not be found");
      return FALSE;
   }

   if (count <= 0) {
      switch (count) {
      case ROADMAP_STREET_NOADDRESS:
         roadmap_dialog_message ("The street address could not be found");
         break;
      case ROADMAP_STREET_NOCITY:
         roadmap_dialog_message ("No city with that name could be found");
         break;
      case ROADMAP_STREET_NOSTREET:
         roadmap_dialog_message ("No street with that name could be found");
         break;
      default:
         roadmap_dialog_message ("The address could not be found");
      }
      return TRUE;
   }

   if (count > 1) {
      /* FIXME: open a dialog that lists all possible streets. */
      roadmap_dialog_message ("Several streets was found");
   }

   line = roadmap_street_get_position (blocks, street_number, &position);

   if (line < 0) {
      roadmap_dialog_message ("The street number could not be found");
      return TRUE;
   }

   gtk_grab_remove (context->dialog);
   gtk_widget_hide (context->dialog);

   context->when_done (&position);

   return FALSE;
}


static GtkWidget *roadmap_dialog_key (RoadMapKey *key, char character) {

   char label[2];

   key->character = character;

   if (character < ' ') {
      if (character == '\r') {
         character = 0xab;
      }
      if (character == '\b') {
         character = '<';
      }
   }
   label[0] = character;
   label[1] = 0;

   key->button    = gtk_button_new_with_label (label);

   gtk_signal_connect (GTK_OBJECT(key->button),
                       "clicked",
                       GTK_SIGNAL_FUNC(roadmap_dialog_key_pressed),
                       key);

   return key->button;
}


static GtkWidget *roadmap_dialog_keyboard (RoadMapAddressDialog *context) {

   int i;
   int j;
   int k;
   char keymap[40] = "1234567890qwertyuiopasdfghjkl\b zxcvbnm \r";

   GtkWidget *keyboard;
   GtkWidget *button;

   keyboard = gtk_table_new (4, 10, TRUE);

   for (i = 0; i < 4; i++) {
      for (j = 0; j < 10; j++) {
         k = (10 * i) + j;
         button = roadmap_dialog_key (&context->keyboard[k], keymap[k]);
         gtk_table_attach_defaults
            (GTK_TABLE(keyboard), button, j, j+1, i, i+1);
         context->keyboard[k].context = context;
      }
   }

   return keyboard;
}


static void roadmap_dialog_address_new (RoadMapAddressDialog *context) {

   int i;
   GtkWidget *ok;
   GtkWidget *cancel;
   GtkWidget *label[4];

   GtkWidget *table;


   context->dialog = gtk_dialog_new();
   gtk_window_set_title (GTK_WINDOW(context->dialog), context->title);

   table = gtk_table_new (3, 2, FALSE);


   label[0] = gtk_label_new ("Number:");
   label[1] = gtk_label_new ("Street:");
   if (context->use_zip) {
      label[2] = gtk_label_new ("Zip:");
   } else {
      label[2] = gtk_label_new ("City:");
   }
   label[3] = gtk_label_new ("State:");

   for (i = 0; i < 4; i++) {

      gtk_table_attach_defaults (GTK_TABLE(table), label[i], 0, 1, i, i+1);

      context->entry[i] = gtk_entry_new ();

      gtk_table_attach (GTK_TABLE(table),
                        context->entry[i], 1, 2, i, i+1,
                        GTK_EXPAND+GTK_FILL+GTK_SHRINK,
                        GTK_EXPAND+GTK_FILL+GTK_SHRINK, 2, 2);

      gtk_signal_connect (GTK_OBJECT(context->entry[i]),
                          "button_press_event",
                          GTK_SIGNAL_FUNC(roadmap_dialog_selected),
                          context);
   }
   
   gtk_box_pack_start (GTK_BOX(GTK_DIALOG(context->dialog)->vbox),
                       table, TRUE, TRUE, 0);

   gtk_box_pack_start (GTK_BOX(GTK_DIALOG(context->dialog)->vbox),
                       roadmap_dialog_keyboard(context), TRUE, TRUE, 0);

   gtk_container_set_border_width
      (GTK_CONTAINER(GTK_BOX(GTK_DIALOG(context->dialog)->vbox)), 4);

   gtk_table_set_row_spacings (GTK_TABLE(table), 3);

   ok = gtk_button_new_with_label ("Ok");
   cancel = gtk_button_new_with_label ("Cancel");

   gtk_signal_connect (GTK_OBJECT(ok),
                       "clicked",
                       GTK_SIGNAL_FUNC(roadmap_dialog_address_ok),
                       context);

   gtk_signal_connect (GTK_OBJECT(cancel),
                       "clicked",
                       GTK_SIGNAL_FUNC(roadmap_dialog_address_cancel),
                       context);

   GTK_WIDGET_SET_FLAGS (ok, GTK_CAN_DEFAULT);
   GTK_WIDGET_SET_FLAGS (cancel, GTK_CAN_DEFAULT);

   gtk_box_pack_start (GTK_BOX(GTK_DIALOG(context->dialog)->action_area),
                       ok, TRUE, TRUE, 0);

   gtk_box_pack_start (GTK_BOX(GTK_DIALOG(context->dialog)->action_area),
                       cancel, TRUE, TRUE, 0);

   gtk_widget_grab_default (ok);

   gtk_grab_add (context->dialog);
}


static void roadmap_dialog_address (RoadMapAddressDialog *context) {

   if (context->dialog != NULL) {

      if (context->dialog->window != NULL) {
         gdk_window_show (context->dialog->window);
         gdk_window_raise (context->dialog->window);
      }

   } else {

      roadmap_dialog_address_new(context);

      gtk_signal_connect (GTK_OBJECT(context->dialog),
                          "destroy",
                          GTK_SIGNAL_FUNC(gtk_widget_destroyed),
                          &(context->dialog));
   }
   gtk_widget_show_all (GTK_WIDGET(context->dialog));
}


static void roadmap_dialog_destination_address_done
               (RoadMapPosition *position) {

    roadmap_canvas_set_destination (position);
}

void roadmap_dialog_destination_address_by_city (void) {

   static RoadMapAddressDialog context = {
      "Destination",
      0,
      roadmap_dialog_destination_address_done
   };

   roadmap_dialog_address (&context);
}


void roadmap_dialog_destination_address_by_zip (void) {

   static RoadMapAddressDialog context = {
      "Destination",
      1,
      roadmap_dialog_destination_address_done
   };

   roadmap_dialog_address (&context);
}


static void roadmap_dialog_location_address_done (RoadMapPosition *position) {

    roadmap_canvas_set_location (position);
}

void roadmap_dialog_location_address_by_city (void) {

   static RoadMapAddressDialog context = {
      "Location",
      0,
      roadmap_dialog_location_address_done
   };

   roadmap_dialog_address (&context);
}

void roadmap_dialog_location_address_by_zip (void) {

   static RoadMapAddressDialog context = {
      "Location",
      1,
      roadmap_dialog_location_address_done
   };

   roadmap_dialog_address (&context);
}

void roadmap_dialog_about (void) {

   roadmap_dialog_message ("RoadMap v0.1\n"
                           "(c) 2002  Pascal Martin\n"
                           "<pascal.martin@iname.com>\n"
                           "Street navigation system");
}

