/* roadmap_math.c - Manage the little math required to place points on a map.
 *
 * LICENSE:
 *
 *   Copyright 2002 Pascal F. Martin
 *
 *   This file is part of RoadMap.
 *
 *   RoadMap is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   RoadMap is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with RoadMap; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * SYNOPSYS:
 *
 *   See roadmap_math.h.
 *
 * These functions are used to compute the position of points on the map,
 * or the distance between two points, given their Tiger position.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "roadmap.h"
#include "roadmap_math.h"
#include "roadmap_square.h"
#include "roadmap_config.h"


#define ROADMAP_BASE_IMPERIAL 0
#define ROADMAP_BASE_METRIC   1


static struct {

   char initialized;

   unsigned short zoom;

   /* The current position shown on the map: */
   RoadMapPosition center;

   /* The center point (current position), in pixel: */
   int center_x;
   int center_y;

   /* The size of the area shown (pixels): */
   int width;
   int height;

   /* The conversion ratio from position to pixels: */
   int zoom_x;
   int zoom_y;

   /* The conversion ratio from position to distance: */

   double unit_per_latitude;  /* Either imperial or metric. */
   double unit_per_longitude;

   double m_per_latitude;
   double m_per_longitude;

   double ft_per_latitude;
   double ft_per_longitude;

   RoadMapPosition topleft;
   RoadMapPosition bottomright;

} RoadMapContext;

double Half_PI;


int RoadmapDefaultZoom = 0;


static void roadmap_math_constants (void) {

   if (RoadmapDefaultZoom == 0) {

      RoadmapDefaultZoom =
          roadmap_config_get_integer ("General", "Default Zoom");

      RoadMapContext.zoom = RoadmapDefaultZoom;
   }

   if (! RoadMapContext.initialized) {

      Half_PI = asin(1);

      RoadMapContext.m_per_latitude  = (1.852 * 60.0) / 1000.0;
      RoadMapContext.ft_per_latitude = 3.2815 * RoadMapContext.m_per_latitude;

      RoadMapContext.initialized = 1;
   }
}


static void roadmap_math_compute_scale (void) {

   int west;
   int east;
   int north;
   int south;

   double ratio;
   GdkPoint point;


   RoadMapContext.center_x = RoadMapContext.width / 2;
   RoadMapContext.center_y = RoadMapContext.height / 2;

   RoadMapContext.zoom_x = RoadMapContext.zoom;
   RoadMapContext.zoom_y = RoadMapContext.zoom;

   /* The horizontal ratio is derived from the vertical one,
    * with a scaling depending on the latitude. The goal is to
    * compute a map projection and avoid an horizontal distortion
    * when getting close to the poles.
    */
   roadmap_square_edges (ROADMAP_SQUARE_GLOBAL, &west, &east, &north, &south);

   ratio = cos((((north + south) / 2) * Half_PI) / 90000000.0);

   RoadMapContext.m_per_longitude = RoadMapContext.m_per_latitude * ratio;
   RoadMapContext.ft_per_longitude = RoadMapContext.ft_per_latitude * ratio;
   RoadMapContext.unit_per_longitude = RoadMapContext.unit_per_latitude * ratio;

   RoadMapContext.zoom_y = (int) (RoadMapContext.zoom_y * ratio);

   RoadMapContext.topleft.longitude =
      RoadMapContext.center.longitude
         - (RoadMapContext.center_x * RoadMapContext.zoom_x);

   RoadMapContext.topleft.latitude =
      RoadMapContext.center.latitude
         + (RoadMapContext.center_y * RoadMapContext.zoom_y);

   point.x = RoadMapContext.width;
   point.y = RoadMapContext.height;
   roadmap_math_to_position (&point, &(RoadMapContext.bottomright));
}


void roadmap_math_initialize (void) {

   roadmap_config_declare_preference ("General", "Default Zoom", "20");
}

void roadmap_math_use_metric (void) {

   roadmap_math_constants ();

   RoadMapContext.unit_per_latitude = RoadMapContext.m_per_latitude;

   roadmap_math_compute_scale ();
}


void roadmap_math_use_imperial (void) {

   roadmap_math_constants ();

   RoadMapContext.unit_per_latitude = RoadMapContext.ft_per_latitude;

   roadmap_math_compute_scale ();
}


int roadmap_math_is_visible (int west, int east, int north, int south) {

   if (west > RoadMapContext.bottomright.longitude ||
       east < RoadMapContext.topleft.longitude ||
       south > RoadMapContext.topleft.latitude ||
       north < RoadMapContext.bottomright.latitude)
   {
       return 0;
   }

   if (west >= RoadMapContext.topleft.longitude &&
       east < RoadMapContext.bottomright.longitude &&
       south > RoadMapContext.bottomright.latitude &&
       north <= RoadMapContext.topleft.latitude)
   {
       return 1;
   }

   return -1;
}


void roadmap_math_zoom_out (void) {

   int zoom;

   roadmap_math_constants ();

   zoom = (3 * RoadMapContext.zoom) / 2;
   if (zoom < 0x10000) {
      RoadMapContext.zoom = (unsigned short) zoom;
   }
   roadmap_math_compute_scale ();
}


void roadmap_math_zoom_in (void) {

   roadmap_math_constants ();

   RoadMapContext.zoom = (2 * RoadMapContext.zoom) / 3;
   if (RoadMapContext.zoom <= 1) {
      RoadMapContext.zoom = 2;
   }
   roadmap_math_compute_scale ();
}


void roadmap_math_zoom_reset (void) {

   roadmap_math_constants ();

   RoadMapContext.zoom = RoadmapDefaultZoom;
   roadmap_math_compute_scale ();
}


int roadmap_math_declutter (int level) {

   return (RoadMapContext.zoom < level);
}


int roadmap_math_thickness (int base) {

   float ratio;

   roadmap_math_constants ();

   ratio = (1.0 * RoadmapDefaultZoom * base) / RoadMapContext.zoom;

   if (ratio < 0.1 / base) {
      return 1;
   }
   if (ratio < 1.0 * base) {
      return base;
   }

   return (int) ratio;
}


void roadmap_math_set_size (int width, int height) {

   roadmap_math_constants ();

   RoadMapContext.width = width;
   RoadMapContext.height = height;

   roadmap_math_compute_scale ();
}


void roadmap_math_set_center (RoadMapPosition *position) {

   roadmap_math_constants ();

   RoadMapContext.center = *position;

   roadmap_math_compute_scale ();
}


void roadmap_math_to_position (GdkPoint *point, RoadMapPosition *position) {

   position->longitude =
      RoadMapContext.topleft.longitude + (point->x * RoadMapContext.zoom_x);

   position->latitude =
      RoadMapContext.topleft.latitude - (point->y * RoadMapContext.zoom_y);
}


void roadmap_math_coordinate (RoadMapPosition *position, GdkPoint *point) {

   point->x = ((position->longitude - RoadMapContext.topleft.longitude)
                 / RoadMapContext.zoom_x);

   point->y = ((RoadMapContext.topleft.latitude - position->latitude)
                 / RoadMapContext.zoom_y);
}


int roadmap_math_distance
       (RoadMapPosition *position1, RoadMapPosition *position2) {

   double x;
   double y;


   x = RoadMapContext.unit_per_longitude
          * (position1->longitude - position2->longitude);
   y = RoadMapContext.unit_per_latitude
          * (position1->latitude  - position2->latitude);

   return (int) sqrt ((x * x) + (y * y));
}


int  roadmap_math_get_distance_from_segment
        (RoadMapPosition *position,
         RoadMapPosition *position1, RoadMapPosition *position2) {

   int distance;
   int minimum;

   double x1;
   double y1;
   double x2;
   double y2;
   double x3;
   double y3;


   x1 = RoadMapContext.unit_per_longitude
           * (position->longitude - position1->longitude);
   y1 = RoadMapContext.unit_per_latitude
           * (position->latitude  - position1->latitude);

   x2 = RoadMapContext.unit_per_longitude
           * (position->longitude - position2->longitude);
   y2 = RoadMapContext.unit_per_latitude
           * (position->latitude  - position2->latitude);

   /* Compute the coordinates of the intersection with the perpendicular. */

   if ((x1 - x2 > -1.0) && (x1 - x2 < 1.0)) {

      /* This approximation is good enough. */
      x3 = (x1 + x2) / 2;
      y3 = 0.0;

   } else {
      
      /* Equation of the line: */

      double a = (y1 - y2) / (x1 - x2);
      double b = y1 - a * x1;

      /* The equation of the perpendicular is: y = - (x / a). */

      x3 = 0.0 - (b / (a + (1.0 / a)));
      y3 = b / ((a * a) + 1.0);
   }


   if ((((x1 >= x3) && (x3 >= x2)) || ((x1 <= x3) && (x3 <= x2))) &&
       (((y1 >= y3) && (y3 >= y2)) || ((y1 <= y3) && (y3 <= y2)))) {

      /* The intersection point is in the segment. */

      return (int) sqrt ((x3 * x3) + (y3 * y3));
   }

   /* The intersection point is not in the segment: use the distance to
    * the closest segment's endpoint.
    */
   minimum  = (int) sqrt ((x1 * x1) + (y1 * y1));
   distance = (int) sqrt ((x2 * x2) + (y2 * y2));

   if (distance < minimum) {
      minimum = distance;
   }

   return minimum;
}


void roadmap_math_screen_edges (int *west, int *east, int *north, int *south) {

   if (west != NULL) {
      *west = RoadMapContext.topleft.longitude;
   }
   if (east != NULL) {
      *east = RoadMapContext.bottomright.longitude;
   }
   if (north != NULL) {
      *north = RoadMapContext.topleft.latitude;
   }
   if (south != NULL) {
      *south = RoadMapContext.bottomright.latitude;
   }
}


int  roadmap_math_street_address (char *image, int length) {

   int i;
   int digit;
   int result = 0;

   for (i = 0; i < length; i++) {

      digit = image[i];

      if ((digit < '0') || (digit > '9')) {

         if (digit == '-' || digit == ' ') {
            continue;
         } else if (digit >= 'A' && digit <= 'Z') {
            digit = '1' + digit - 'A';
         } else if (digit >= 'a' && digit <= 'z') {
            digit = '1' + digit - 'a';
         } else  {
            roadmap_log (ROADMAP_ERROR, "bad numerical character %c", digit);
            continue;
         }
      }
      result = (result * 10) + (digit - '0');
   }

   return result;
}

