/* roadmap_preferences.c - handle the roadmap dialogs managing user preferences.
 *
 * LICENSE:
 *
 *   Copyright 2002 Pascal F. Martin
 *
 *   This file is part of RoadMap.
 *
 *   RoadMap is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   RoadMap is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with RoadMap; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * SYNOPSYS:
 *
 *   void roadmap_preferences_edit (void);
 */

#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "roadmap.h"
#include "roadmap_types.h"
#include "roadmap_gui.h"
#include "roadmap_preferences.h"


typedef struct EditableItemRecord EditableItem;
typedef struct SelectableItemRecord SelectableItem;


struct EditableItemRecord {

   EditableItem *next;

   int   type;
   char *category;
   char *name;

   SelectableItem *selection;

   union {
      GtkWidget *entry;
      SelectableItem *selected;
   } value;

   GtkWidget *dialog;

};

struct SelectableItemRecord {

   SelectableItem *next;
   EditableItem   *parent;
   char           *text;
};


static gint roadmap_preferences_cancel (GtkWidget *w, gpointer data) {

   EditableItem *next;
   EditableItem *item = (EditableItem *) data;
   SelectableItem *selection;
   SelectableItem *next_selection;
   GtkWidget *dialog;

   if (item != NULL) {

      dialog = item->dialog;

      for (item = (EditableItem *) data; item != NULL; item = next) {

         for (selection = item->selection;
              selection != NULL;
              selection = next_selection) {

            next_selection = selection->next;
            free (selection);
         }

         next = item->next;
         free (item);
      }

      gtk_grab_remove (dialog);
      gtk_widget_destroy (dialog);
   }

   return FALSE;
}

static gint roadmap_preferences_ok (GtkWidget *w, gpointer data) {

   EditableItem *item;

   for (item = (EditableItem *) data; item != NULL; item = item->next) {

      switch (item->type) {

      case ROADMAP_CONFIG_ENUM:

         if (item->value.selected != NULL) {
            if (item->value.selected->text != NULL) {
               roadmap_config_set
                  (item->category, item->name, item->value.selected->text);
            }
         }
         break;

      case ROADMAP_CONFIG_COLOR:
      case ROADMAP_CONFIG_STRING:

         roadmap_config_set
            (item->category,
             item->name,
             gtk_entry_get_text (GTK_ENTRY(item->value.entry)));
      }
   }

   return roadmap_preferences_cancel (w, data);
}

static gint roadmap_preferences_selected (gpointer data, GtkMenuItem *w) {

   SelectableItem *selection = (SelectableItem *)data;

   if (selection != NULL) {
      selection->parent->value.selected = selection;
   }

   return FALSE;
}


static void roadmap_preferences_new (char *title, void *cursor) {

   void *next;
   char full_title[256];

   int i;
   int count = 0;
   int lines[256];
   char *categories[256];
   GtkWidget *pages[256];

   int   type;
   char *category;
   char *name;
   char *value;

   GtkWidget *ok;
   GtkWidget *cancel;
   GtkWidget *label;
   GtkWidget *item_widget;
   GtkWidget *dialog;
   GtkWidget *notebook;

   EditableItem *new_item;
   EditableItem *last_item = NULL;


   if (cursor == NULL) return;

   dialog = gtk_dialog_new();

   roadmap_window_set_size (dialog, 400, 300);

   strcpy (full_title, "RoadMap ");
   strcat (full_title, title);
   gtk_window_set_title (GTK_WINDOW(dialog), full_title);

   notebook = gtk_notebook_new();

   gtk_notebook_set_scrollable (GTK_NOTEBOOK(notebook), TRUE);

   gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dialog)->vbox),
                       notebook, TRUE, TRUE, 0);

   while (cursor != NULL) {

      type = roadmap_config_get_type (cursor);
      next = roadmap_config_scan (cursor, &category, &name, &value);

      for (i = count - 1; i >= 0; i--) {
         if (strcmp (category, categories[i]) == 0) break;
      }

      if (i < 0) {

         /* This is a new category. */

         if (count >= 256) continue; /* Cannot handle more category. */

         label = gtk_label_new (category);

         lines[count] = 0;
         pages[count] = gtk_table_new (1, 2, FALSE);
         categories[count] = category;

         i = count;
         count += 1;

         gtk_notebook_append_page
            (GTK_NOTEBOOK(notebook), pages[i], label);

      } else {
         gtk_table_resize (GTK_TABLE(pages[i]), lines[i]+2, 2);
      }

      label = gtk_label_new (name);
      gtk_label_set_justify (GTK_LABEL(label), GTK_JUSTIFY_LEFT);
      gtk_label_set_text    (GTK_LABEL(label), name); // ?? does not justify ?

      gtk_table_attach (GTK_TABLE(pages[i]),
                        label, 0, 1, lines[i], lines[i]+1,
                        GTK_FILL, GTK_FILL, 2, 2);

      new_item = malloc (sizeof(EditableItem));
      if (new_item == NULL) {
         roadmap_log (ROADMAP_FATAL, "no more memory");
      }

      new_item->type      = type;
      new_item->category  = category;
      new_item->name      = name;
      new_item->dialog    = dialog;
      new_item->selection = NULL;

      switch (type) {

      case ROADMAP_CONFIG_ENUM:

         item_widget = gtk_option_menu_new ();

         {
            char *text;
            SelectableItem *choice;
            GtkWidget *menu_item;
            GtkWidget *menu = gtk_menu_new ();
            void *enumeration = roadmp_config_get_enumeration (cursor);

            char *current_value = roadmap_config_get (category, name);


            for (enumeration = roadmp_config_get_enumeration (cursor);
                 enumeration != NULL;
                 enumeration = roadmp_config_get_enumeration_next (enumeration)) {
               text = roadmp_config_get_enumeration_value (enumeration);

               choice = malloc (sizeof(SelectableItem));
               if (choice == NULL) {
                  roadmap_log (ROADMAP_FATAL, "no more memory");
               }

               choice->parent = new_item;
               choice->text   = text;
               choice->next   = new_item->selection;

               new_item->selection = choice;

               menu_item = gtk_menu_item_new_with_label (text);

               if (strcmp (text, current_value) == 0) {
                  gtk_menu_prepend (GTK_MENU(menu), menu_item);
               } else {
                  gtk_menu_append (GTK_MENU(menu), menu_item);
               }

               gtk_signal_connect_object
                  (GTK_OBJECT(menu_item),
                   "activate",
                   GTK_SIGNAL_FUNC(roadmap_preferences_selected),
                   (gpointer)choice);

               gtk_widget_show (menu_item);
            }

            gtk_option_menu_set_menu (GTK_OPTION_MENU(item_widget), menu);
         }

         new_item->value.selected = NULL;
         break;

      case ROADMAP_CONFIG_COLOR:
      case ROADMAP_CONFIG_STRING:

         item_widget = gtk_entry_new ();
         gtk_entry_set_text (GTK_ENTRY(item_widget), value);

         new_item->value.entry = item_widget;
         break;

      default:
         roadmap_log (ROADMAP_FATAL, "invalid preference item type %d", type);
         return;
      }

      gtk_table_attach (GTK_TABLE(pages[i]),
                        item_widget, 1, 2, lines[i], lines[i]+1,
                        GTK_EXPAND+GTK_FILL+GTK_SHRINK,
                        GTK_EXPAND+GTK_FILL+GTK_SHRINK, 2, 2);


      new_item->next = last_item;
      last_item = new_item;

      lines[i] += 1;

      cursor = next;
   }

   if (last_item == NULL) {
      roadmap_log (ROADMAP_ERROR, "no preference item registered");
   }

   gtk_container_set_border_width
      (GTK_CONTAINER(GTK_BOX(GTK_DIALOG(dialog)->vbox)), 4);


   ok     = gtk_button_new_with_label ("Ok");
   cancel = gtk_button_new_with_label ("Cancel");

   GTK_WIDGET_SET_FLAGS (ok, GTK_CAN_DEFAULT);
   GTK_WIDGET_SET_FLAGS (cancel, GTK_CAN_DEFAULT);

   gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dialog)->action_area),
                       ok, TRUE, TRUE, 0);

   gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dialog)->action_area),
                       cancel, TRUE, TRUE, 0);

   gtk_signal_connect (GTK_OBJECT(ok),
                       "clicked",
                       GTK_SIGNAL_FUNC(roadmap_preferences_ok),
                       last_item);

   gtk_signal_connect (GTK_OBJECT(cancel),
                       "clicked",
                       GTK_SIGNAL_FUNC(roadmap_preferences_cancel),
                       last_item);

   gtk_widget_grab_default (ok);

   gtk_grab_add (dialog);

   gtk_widget_show_all (GTK_WIDGET(dialog));
}


void roadmap_preferences_edit (void) {
   roadmap_preferences_new ("Preferences", roadmap_config_first_preference());
}

