/* roadmap_square.c - Manage a county area, divided in small squares.
 *
 * LICENSE:
 *
 *   Copyright 2002 Pascal F. Martin
 *
 *   This file is part of RoadMap.
 *
 *   RoadMap is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   RoadMap is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with RoadMap; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * SYNOPSYS:
 *
 *   See roadmap_square.h.
 *
 * These functions are used to retrieve the squares that make the county
 * area. A special square (ROADMAP_SQUARE_GLOBAL) is used to describe the
 * global county area (vs. a piece of it).
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "roadmap.h"
#include "roadmap_dbread.h"
#include "roadmap_db_square.h"

#include "roadmap_square.h"

static char *RoadMapSquareType = "RoadMapSquareContext";

typedef struct {

   int index;
   RoadMapSquare *square;

} RoadMapSquareGrid;

typedef struct {

   char *type;

   RoadMapGlobal *SquareGlobal;
   RoadMapSquare *Square;

   RoadMapSquareGrid *SquareGrid;
   int                SquareGridCount;

} RoadMapSquareContext;

static RoadMapSquareContext *RoadMapSquareActive = NULL;


static void *roadmap_square_map (roadmap_db *root) {

   RoadMapSquareContext *context;

   int i;
   int count;
   roadmap_db *global_table;
   roadmap_db *square_table;


   context = malloc(sizeof(RoadMapSquareContext));
   if (context == NULL) {
      roadmap_log(ROADMAP_FATAL, "no more memory");
   }
   context->type = RoadMapSquareType;

   global_table  = roadmap_db_get_subsection (root, "global");
   square_table = roadmap_db_get_subsection (root, "data");

   context->SquareGlobal = (RoadMapGlobal *) roadmap_db_get_data (global_table);
   context->Square       = (RoadMapSquare *) roadmap_db_get_data (square_table);

   count = context->SquareGlobal->count_longitude
              * context->SquareGlobal->count_latitude;

   context->SquareGrid =
      (RoadMapSquareGrid *) calloc (sizeof(RoadMapSquare), count);
   if (context->SquareGrid == NULL) {
      roadmap_log (ROADMAP_FATAL, "no more memory");
   }
   context->SquareGridCount = count;

   for (i = count - 1; i >= 0; i--) {
      context->SquareGrid[i].index = -1;
   }

   for (i = context->SquareGlobal->count_squares - 1; i >= 0; i--) {

      RoadMapSquare *this_square = context->Square + i;
      RoadMapSquareGrid *grid = context->SquareGrid + this_square->position;

      grid->index = i;
      grid->square = this_square;
   }

   return context;
}

static void roadmap_square_activate (void *context) {

   RoadMapSquareContext *square_context = (RoadMapSquareContext *) context;

   if (square_context->type != RoadMapSquareType) {
      roadmap_log(ROADMAP_FATAL, "cannot unmap (bad context type)");
   }
   RoadMapSquareActive = square_context;
}

static void roadmap_square_unmap (void *context) {

   RoadMapSquareContext *square_context = (RoadMapSquareContext *) context;

   if (square_context->type != RoadMapSquareType) {
      roadmap_log(ROADMAP_FATAL, "cannot unmap (bad context type)");
   }
   if (RoadMapSquareActive == square_context) {
      RoadMapSquareActive = NULL;
   }
   free(square_context->SquareGrid);
   free(square_context);
}

roadmap_db_handler RoadMapSquareHandler = {
   "square",
   roadmap_square_map,
   roadmap_square_activate,
   roadmap_square_unmap
};



int roadmap_square_count (void) {

   return RoadMapSquareActive->SquareGridCount;
}


static int roadmap_square_on_grid (RoadMapPosition *position) {

   int x;
   int y;

   RoadMapGlobal *global = RoadMapSquareActive->SquareGlobal;


   x = (position->longitude - global->min_longitude) / global->step_longitude;
   if (x < 0 || x > global->count_longitude) {
      return -1;
   }

   y = (position->latitude  - global->min_latitude)  / global->step_latitude;
   if (y < 0 || y > global->count_latitude) {
      return -1;
   }

   if (x >= global->count_longitude) {
      x = global->count_longitude - 1;
   }
   if (y >= global->count_latitude) {
      y = global->count_latitude - 1;
   }
   return (x * global->count_latitude) + y;
}


static int roadmap_square_location (RoadMapPosition *position) {

   int square;

   RoadMapSquareGrid *grid = RoadMapSquareActive->SquareGrid;

   square = roadmap_square_on_grid (position);
   if (square < 0) {
      return -1;
   }

   /* The computation above may have rouding errors: adjust. */

   while (grid[square].square == NULL ||
          grid[square].square->max_longitude <= position->longitude) {
      square++;
      if (square >= RoadMapSquareActive->SquareGridCount) return -1;
   }
   while (grid[square].square == NULL ||
          grid[square].square->min_longitude > position->longitude) {
      square--;
      if (square < 0) return -1;
   }
   while (grid[square].square == NULL ||
          grid[square].square->max_latitude <= position->latitude) {
      square++;
      if (square >= RoadMapSquareActive->SquareGridCount) return -1;
   }
   while (grid[square].square == NULL ||
          grid[square].square->min_latitude > position->latitude) {
      square--;
      if (square < 0) return -1;
   }

   return square;
}


int roadmap_square_search (RoadMapPosition *position) {

   int square = roadmap_square_location (position);

   RoadMapSquareGrid *grid = RoadMapSquareActive->SquareGrid;

   if (square < 0) {
      return ROADMAP_SQUARE_OTHER;
   }

   if ((grid[square].square == NULL) ||
       (grid[square].square->min_longitude > position->longitude) ||
       (grid[square].square->max_longitude < position->longitude) ||
       (grid[square].square->min_latitude > position->latitude) ||
       (grid[square].square->max_latitude < position->latitude)) {

      return ROADMAP_SQUARE_GLOBAL;
   }

   return square;
}


void  roadmap_square_min (int square, RoadMapPosition *position) {

   RoadMapGlobal *global = RoadMapSquareActive->SquareGlobal;


   /* Default values. */
   position->longitude = global->min_longitude;
   position->latitude  = global->min_latitude;

   if (square == ROADMAP_SQUARE_GLOBAL) {
      return;
   }

   if ((square < 0) || (square >= RoadMapSquareActive->SquareGridCount)) {
      roadmap_log (ROADMAP_FATAL, "invalid square index %d", square);
   }

   square = RoadMapSquareActive->SquareGrid[square].index;
   if (square < 0) {
      return;
   }
   position->longitude = RoadMapSquareActive->Square[square].min_longitude;
   position->latitude  = RoadMapSquareActive->Square[square].min_latitude;
}


void  roadmap_square_edges
         (int square, int *west, int *east, int *north, int *south) {


   if (RoadMapSquareActive == NULL) {

      *west = 0;
      *east = 0;
      *north = 0;
      *south = 0;

      return;
   }

   if (square == ROADMAP_SQUARE_GLOBAL) {

      RoadMapGlobal *global = RoadMapSquareActive->SquareGlobal;

      *west = global->min_longitude;
      *east = global->max_longitude;
      *north = global->max_latitude;
      *south = global->min_latitude;

      return;
   }

   if ((square < 0) || (square >= RoadMapSquareActive->SquareGridCount)) {
      roadmap_log (ROADMAP_FATAL, "invalid square index %d", square);
   }
   square = RoadMapSquareActive->SquareGrid[square].index;

   {
      RoadMapSquare *square_item = RoadMapSquareActive->Square + square;

      *west =  square_item->min_longitude;
      *east =  square_item->max_longitude;
      *north = square_item->max_latitude;
      *south = square_item->min_latitude;
   }
}


static int roadmap_square_neighborhood_y
              (int x, int y, int *neighbors, char *included) {

   int count = 0;
   int adjacent;

   if (y > 0) {

      adjacent = (x * RoadMapSquareActive->SquareGlobal->count_latitude)
                     + (y - 1);

      if (adjacent < 0 || adjacent >= RoadMapSquareActive->SquareGridCount) {

         roadmap_log (ROADMAP_ERROR, "invalid grid index %d", adjacent);

      } else if (included[adjacent] == 0) {

         neighbors[count++] = adjacent;
         included[adjacent] = 1;
      }
   }

   if (y < RoadMapSquareActive->SquareGlobal->count_latitude - 1) {

      adjacent = (x * RoadMapSquareActive->SquareGlobal->count_latitude)
                     + (y + 1);

      if (adjacent < 0 || adjacent >= RoadMapSquareActive->SquareGridCount) {

         roadmap_log (ROADMAP_ERROR, "invalid grid index %d", adjacent);

      } else if (included[adjacent] == 0) {

         neighbors[count++] = adjacent;
         included[adjacent] = 1;
      }
   }

   return count;
}


static int roadmap_square_neighborhood_x
              (int x, int y, int *neighbors, char *included) {

   int count = 0;
   int adjacent;

   adjacent = (x * RoadMapSquareActive->SquareGlobal->count_latitude) + y;

   if (adjacent < 0 || adjacent >= RoadMapSquareActive->SquareGridCount) {

      roadmap_log (ROADMAP_ERROR, "invalid grid index %d", adjacent);

   } else if (included[adjacent] == 0) {

      neighbors[count++] = adjacent;
      included[adjacent] = 1;
   }

   count += roadmap_square_neighborhood_y (x, y, neighbors + count, included);

   return count;
}


static int roadmap_square_neighborhood_xy
              (int grid, int *neighbors, char *included) {

   int x;
   int y;
   int count = 0;

   RoadMapGlobal *global = RoadMapSquareActive->SquareGlobal;


   x = grid / global->count_latitude;
   y = grid % global->count_latitude;

   if (x > 0) {

      count += roadmap_square_neighborhood_x
                  (x - 1, y, neighbors + count, included);
   }

   if (x < global->count_longitude - 1) {

      count += roadmap_square_neighborhood_x
                  (x + 1, y, neighbors + count, included);
   }

   count += roadmap_square_neighborhood_y (x, y, neighbors + count, included);

   return count;
}


int roadmap_square_neighborhood (int square, int *neighbors) {

   int count;
   char *included;


   if (square < 0 || square >= RoadMapSquareActive->SquareGridCount) {
      roadmap_log (ROADMAP_FATAL, "invalid square index %d", square);
   }

   included = calloc (RoadMapSquareActive->SquareGridCount, 1);

   count = roadmap_square_neighborhood_xy (square, neighbors, included);

   free (included);

   return count;
}


static int roadmap_square_search_around
              (int square, int *neighbors, int size,
               char *checked, char *included);

static int roadmap_square_search_around_x
              (int square, int *neighbors, int size,
               char *checked, char *included) {

   RoadMapGlobal *global = RoadMapSquareActive->SquareGlobal;

   int count = 0;
   int x = square / global->count_latitude;


   if (x < global->count_longitude - 1) {

      if (count >= size - 8) {
         return count;
      }
      count += roadmap_square_search_around
                  (square + global->count_latitude,
                   neighbors+count,
                   size-count,
                   checked, included);
   }

   if (x > 0) {

      if (count >= size - 8) {
         return count;
      }
      count += roadmap_square_search_around
                  (square - global->count_latitude,
                   neighbors+count,
                   size-count,
                   checked, included);
   }

   return count;
}


static int roadmap_square_search_around
              (int square, int *neighbors, int size,
               char *checked, char *included) {

   int y;
   int count;
   int count_latitude;


   if (square < 0 || square >= RoadMapSquareActive->SquareGridCount) {
      roadmap_log (ROADMAP_ERROR,
                   "invalid square %d (valid range is 0..%d)",
                   square,
                   RoadMapSquareActive->SquareGridCount);
      return 0;
   }

   if (checked[square]) return 0;
   checked[square] = 1;

   count = roadmap_square_neighborhood_xy (square, neighbors, included);

   if (count >= size - 8) {
      return count;
   }
   count += roadmap_square_search_around_x
               (square, neighbors+count, size-count, checked, included);

   count_latitude = RoadMapSquareActive->SquareGlobal->count_latitude;

   y = square % count_latitude;

   if (y < count_latitude - 1) {

      if (count >= size - 8) {
         return count;
      }
      count += roadmap_square_search_around
                  (square+1, neighbors+count, size-count, checked, included);
   }

   if (y > 0) {

      if (count >= size - 8) {
         return count;
      }
      count += roadmap_square_search_around
                  (square-1, neighbors+count, size-count, checked, included);
   }

   return count;
}


int roadmap_square_closest
       (RoadMapPosition *position, int *neighbors, int size) {

   int count;
   int square;
   RoadMapPosition adjusted;

   char *checked;
   char *included;

   RoadMapGlobal *global = RoadMapSquareActive->SquareGlobal;


   if (position->longitude < global->min_longitude) {
      adjusted.longitude = global->min_longitude;
   } else if (position->longitude > global->max_longitude) {
      adjusted.longitude = global->max_longitude;
   } else {
      adjusted.longitude = position->longitude;
   }

   if (position->latitude < global->min_latitude) {
      adjusted.latitude = global->min_latitude;
   } else if (position->latitude > global->max_latitude) {
      adjusted.latitude = global->max_latitude;
   } else {
      adjusted.latitude = position->latitude;
   }

   square = roadmap_square_on_grid (&adjusted);
   if (square < 0) {
      return -1;
   }

   checked  = calloc (RoadMapSquareActive->SquareGridCount, 1);
   included = calloc (RoadMapSquareActive->SquareGridCount, 1);

   count = roadmap_square_search_around
              (square, neighbors, size, checked, included);

   free (checked);
   free (included);

   return count;
}


int roadmap_square_index (int square) {

   if (square < 0 || square >= RoadMapSquareActive->SquareGridCount) {
      roadmap_log (ROADMAP_FATAL, "invalid square index %d", square);
   }

   return RoadMapSquareActive->SquareGrid[square].index;
}


int roadmap_square_from_index (int index) {

   if (RoadMapSquareActive == NULL) {
      return -1;
   }

   if (index < 0 ||
       index >= RoadMapSquareActive->SquareGlobal->count_squares) {
      return -1;
   }

   return RoadMapSquareActive->Square[index].position;
}

