/* roadmap_menus.c - Create and manage the menus and toolbar.
 *
 * LICENSE:
 *
 *   Copyright 2002 Pascal F. Martin
 *
 *   This file is part of RoadMap.
 *
 *   RoadMap is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   RoadMap is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with RoadMap; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * SYNOPSYS:
 *
 *   void roadmap_window_new (void);
 *   void roadmap_window_main_loop (void);
 */

#include <stdio.h>

#include "roadmap_gui.h"
#include "roadmap_dialog.h"
#include "roadmap_config.h"
#include "roadmap_preferences.h"

#include <gdk/gdkkeysyms.h>


typedef void (*RoadMapMoveMap) (void);
typedef void (*RoadMapGotoMap) (void *);


static void roadmap_window_close (GtkWidget *widget, gpointer data) {

   roadmap_canvas_close();
   gtk_main_quit();
}

static void roadmap_menu_exit(gpointer data, guint action, GtkWidget *widget) {

   roadmap_canvas_close();
   roadmap_config_update();
   gtk_main_quit();
}

static void roadmap_menu_preferences
               (gpointer data, guint action, GtkWidget *widget) {
   roadmap_preferences_edit();
}

static void roadmap_menu_destination_by_city
               (gpointer data, guint action, GtkWidget *widget) {

   roadmap_dialog_destination_address_by_city ();
}

static void roadmap_menu_destination_by_zip
               (gpointer data, guint action, GtkWidget *widget) {

   roadmap_dialog_destination_address_by_zip ();
}

static void roadmap_menu_goto_by_city
               (gpointer data, guint action, GtkWidget *widget) {

   roadmap_dialog_location_address_by_city ();
}

static void roadmap_menu_goto_by_zip
               (gpointer data, guint action, GtkWidget *widget) {

   roadmap_dialog_location_address_by_zip ();
}

static void roadmap_menu_about (GtkWidget *widget, gpointer data) {
   roadmap_dialog_about();
}

static void roadmap_tool_move (GtkWidget *widget, gpointer data) {

   (*((RoadMapMoveMap)data)) ();
}


static void roadmap_tool_goto (GtkWidget *widget, gpointer data) {

   (*((RoadMapGotoMap)data)) (NULL);
}


static gint roadmap_window_key_pressed (GtkWidget *w, GdkEventKey *event) {

   switch (event->keyval) {

      case GDK_Left:   roadmap_canvas_move_west();    break;
      case GDK_Right:  roadmap_canvas_move_east();    break;
      case GDK_Up:     roadmap_canvas_move_north();   break;
      case GDK_Down:   roadmap_canvas_move_south();   break;

      /* These binding are for the iPAQ buttons: */
      case 0x1008ff1a: roadmap_canvas_zoom_reset();   break;
      case 0x1008ff20: roadmap_canvas_zoom_in();      break;
      case 0xaf9:      roadmap_canvas_zoom_out();     break;
      case 0xff67:     roadmap_menu_exit(NULL, 0, w); break;

      /* These binding are for regular keyboards: */
      case GDK_Q:
      case GDK_q:      roadmap_menu_exit(NULL, 0, w); break;
      case GDK_R:
      case GDK_r:      roadmap_canvas_zoom_reset();   break;
      case GDK_plus:   roadmap_canvas_zoom_in();      break;
      case GDK_minus:  roadmap_canvas_zoom_out();     break;
   }

   return FALSE;
}


static GtkItemFactoryEntry RoadMap_Menus[] = {

   {"/_File",                    NULL,          NULL,  0, "<Branch>"},
   {"/File/_Preferences",        NULL,          roadmap_menu_preferences, 0},
   {"/File/",                    NULL,          NULL,  0, "<Separator>"},
   {"/File/_Quit",               "<control>Q",  roadmap_menu_exit,        0},
   {"/_Locations",               NULL,          NULL,  0, "<Branch>"},
   {"/Locations/_Destination By City", "<control>D", roadmap_menu_destination_by_city, 0},
#if 0
   {"/Locations/_Destination By Zip",  "<control>Z", roadmap_menu_destination_by_zip, 0},
#endif
   {"/Locations/_Location By City",    "<control>P", roadmap_menu_goto_by_city,        0},
#if 0
   {"/Locations/_Location By Zip",     "<control>X", roadmap_menu_goto_by_zip,        0},
#endif
   {"/_Screen",                  NULL,          NULL,  0, "<Branch>"},
   {"/Screen/Zoom _In",          NULL,          roadmap_tool_move, (guint)roadmap_canvas_zoom_in},
   {"/Screen/Zoom _Out",         NULL,          roadmap_tool_move, (guint)roadmap_canvas_zoom_out},
   {"/Screen/Zoom _Reset",       NULL,          roadmap_tool_move, (guint)roadmap_canvas_zoom_reset},
   {"/Screen/",                  NULL,          NULL,  0, "<Separator>"},
   {"/Screen/_North",            NULL,          roadmap_tool_move, (guint)roadmap_canvas_move_north},
   {"/Screen/_West",             NULL,          roadmap_tool_move, (guint)roadmap_canvas_move_west},
   {"/Screen/_East",             NULL,          roadmap_tool_move, (guint)roadmap_canvas_move_east},
   {"/Screen/_South",            NULL,          roadmap_tool_move, (guint)roadmap_canvas_move_south},
   {"/Screen/",                  NULL,          NULL,  0, "<Separator>"},
   {"/Screen/Go to _Destination",NULL,          roadmap_tool_goto, (guint)roadmap_canvas_set_destination},
   {"/Screen/Go to _Location",   NULL,          roadmap_tool_goto, (guint)roadmap_canvas_set_location},
   {"/_Help",                    NULL,          NULL,  0, "<Branch>"},
   {"/Help/_About",              NULL,          roadmap_menu_about,       0},
};

static int RoadMap_Menus_Count =
               sizeof(RoadMap_Menus) / sizeof(RoadMap_Menus[0]);


static GtkWidget *roadmap_toolbar_new (void) {

   GtkWidget *toolbar;

   toolbar = gtk_toolbar_new (GTK_ORIENTATION_HORIZONTAL,
                              GTK_TOOLBAR_TEXT);

   gtk_toolbar_append_item (GTK_TOOLBAR(toolbar),
                            "+", "Zoom into the map", NULL,
                            NULL,
                            roadmap_tool_move, roadmap_canvas_zoom_in);

   gtk_toolbar_append_item (GTK_TOOLBAR(toolbar),
                            "-", "Zoom out from the map", NULL,
                            NULL,
                            roadmap_tool_move, roadmap_canvas_zoom_out);

   gtk_toolbar_append_item (GTK_TOOLBAR(toolbar),
                            "R", "Reset zoom to default", NULL,
                            NULL,
                            roadmap_tool_move, roadmap_canvas_zoom_reset);

   gtk_toolbar_append_space (GTK_TOOLBAR(toolbar));

   gtk_toolbar_append_item (GTK_TOOLBAR(toolbar),
                            "N", "Move the map to the north", NULL,
                            NULL,
                            roadmap_tool_move, roadmap_canvas_move_north);

   gtk_toolbar_append_item (GTK_TOOLBAR(toolbar),
                            "W", "Move the map to the west", NULL,
                            NULL,
                            roadmap_tool_move, roadmap_canvas_move_west);

   gtk_toolbar_append_item (GTK_TOOLBAR(toolbar),
                            "E", "Move the map to the east", NULL,
                            NULL,
                            roadmap_tool_move, roadmap_canvas_move_east);

   gtk_toolbar_append_item (GTK_TOOLBAR(toolbar),
                            "S", "Move the map to the south", NULL,
                            NULL,
                            roadmap_tool_move, roadmap_canvas_move_south);

   gtk_toolbar_append_space (GTK_TOOLBAR(toolbar));

   gtk_toolbar_append_item (GTK_TOOLBAR(toolbar),
                            "D", "center the map on the destination", NULL,
                            NULL,
                            roadmap_tool_goto, roadmap_canvas_set_destination);

   gtk_toolbar_append_item (GTK_TOOLBAR(toolbar),
                            "L", "center the map on the current location", NULL,
                            NULL,
                            roadmap_tool_goto, roadmap_canvas_set_location);

   return toolbar;
}


void roadmap_window_initialize (void) {

   roadmap_config_declare_enumeration_preference
      ("General", "Toolbar", "yes", "no", NULL);
}


void roadmap_window_set_size (GtkWidget *w, int width, int height) {

   int screen_width = gdk_screen_width();

   if (screen_width <= width - 10) {

      /* Small screen: take it all (almost: keep room for the wm). */

      gtk_widget_set_usize (w, screen_width-10, gdk_screen_height()-40);

   } else {

      gtk_widget_set_usize (w, width, height);
   }
}


void roadmap_window_new (void) {

   GtkWidget      *window;
   GtkWidget      *outer_box;
   GtkAccelGroup  *accel_group;
   GtkItemFactory *item_factory;

   char *use_toolbar = roadmap_config_get ("General", "Toolbar");


   window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

   gtk_widget_set_events (window, GDK_KEY_PRESS_MASK);

   roadmap_window_set_size (window, 800, 600);

   gtk_signal_connect (GTK_OBJECT(window), "destroy",
                       GTK_SIGNAL_FUNC(gtk_widget_destroyed),
                       &window);

   gtk_signal_connect (GTK_OBJECT(window), "delete_event",
                       GTK_SIGNAL_FUNC(roadmap_window_close),
                       NULL);

   gtk_signal_connect (GTK_OBJECT(window), "key_press_event",
                       GTK_SIGNAL_FUNC(roadmap_window_key_pressed),
                       NULL);

   accel_group = gtk_accel_group_new();

   item_factory = gtk_item_factory_new
                     (GTK_TYPE_MENU_BAR, "<blah>", accel_group);

   gtk_item_factory_create_items (item_factory,
                                  RoadMap_Menus_Count, RoadMap_Menus, NULL);

   gtk_accel_group_attach (accel_group, GTK_OBJECT(window));

   gtk_window_set_title (GTK_WINDOW(window), "RoadMap");
   gtk_container_border_width (GTK_CONTAINER(window), 0);

   outer_box = gtk_vbox_new (FALSE, 0);
   gtk_container_add (GTK_CONTAINER(window), outer_box);

   gtk_box_pack_start (GTK_BOX(outer_box),
                       gtk_item_factory_get_widget (item_factory, "<blah>"),
                       FALSE, FALSE, 0);

   if (strcmp (use_toolbar, "yes") == 0) {

      gtk_box_pack_start (GTK_BOX(outer_box),
                          roadmap_toolbar_new (),
                          FALSE, FALSE, 0);
   }

   gtk_box_pack_start (GTK_BOX(outer_box),
                       roadmap_canvas (),
                       TRUE, TRUE, 2);

   gtk_widget_show_all (window);
}


void roadmap_window_main_loop (void) {
   gtk_main();
}

